"""
transfer matrix related functions

A collection of functions to compute 4x4 and 6x6 transfer matrices
"""
import numpy
from ..lattice import Lattice, Element, DConstant, Refpts, Orbit
from ..lattice import frequency_control, get_uint32_index
from ..lattice.elements import Dipole, M66
from ..tracking import internal_lpass, internal_epass
from .orbit import find_orbit4, find_orbit6
from .amat import jmat, symplectify

__all__ = ['find_m44', 'find_m66', 'find_elem_m66', 'gen_m66_elem']

_jmt = jmat(2)


def find_m44(ring: Lattice, dp: float = None, refpts: Refpts = None,
             dct: float = None, df: float = None,
             orbit: Orbit = None, keep_lattice: bool = False, **kwargs):
    """One turn 4x4 transfer matrix

    :py:func:`find_m44` finds the 4x4 transfer matrix of an accelerator
    lattice by differentiation of trajectories near the closed orbit.

    Important:
        :py:func:`find_m44` assumes constant momentum deviation.
        The ``PassMethod`` used for any element **SHOULD NOT**:

        1.  change the longitudinal momentum dP
            (cavities , magnets with radiation, ...)
        2.  have any time dependence (localized impedance, fast kickers, ...)

    Unless an input orbit is introduced, :py:func:`find_m44` assumes that the
    lattice is a ring and first finds the closed orbit.

    Parameters:
        ring:           Lattice description (radiation must be OFF)
        dp:             Momentum deviation.
        refpts:         Observation points
        dct:            Path lengthening.
        df:             Deviation of RF frequency.
        orbit:          Avoids looking for initial the closed orbit if it is
          already known ((6,) array).
        keep_lattice:   Assume no lattice change since the previous tracking.
          Default: :py:obj:`False`

    Keyword Args:
        full (bool):    If :py:obj:`True`, matrices are full 1-turn matrices
          at the entrance of each element indexed by refpts. If :py:obj:`False`
          (Default), matrices are between the entrance of the first element and
          the entrance of the selected element
        XYStep (float): Step size.
          Default: :py:data:`DConstant.XYStep <.DConstant>`

    Returns:
        m44:    full one-turn matrix at the entrance of the first element
        ms:     4x4 transfer matrices between the entrance of the first
          element and each element indexed by refpts: (Nrefs, 4, 4) array

    See also:
         :py:func:`find_m66`, :py:func:`.find_orbit4`
    """

    def mrotate(m):
        m = numpy.squeeze(m)
        return m.dot(m44.dot(_jmt.T.dot(m.T.dot(_jmt))))

    xy_step = kwargs.pop('XYStep', DConstant.XYStep)
    full = kwargs.pop('full', False)
    if orbit is None:
        orbit, _ = find_orbit4(ring, dp=dp, dct=dct, df=df,
                               keep_lattice=keep_lattice, XYStep=xy_step)
        keep_lattice = True
    # Construct matrix of plus and minus deltas
    # scaling = 2*xy_step*numpy.array([1.0, 0.1, 1.0, 0.1])
    scaling = xy_step * numpy.array([1.0, 1.0, 1.0, 1.0])
    dg = numpy.asfortranarray(
        numpy.concatenate((0.5 * numpy.diag(scaling), numpy.zeros((2, 4)))))
    dmat = numpy.concatenate((dg, -dg), axis=1)
    # Add the deltas to multiple copies of the closed orbit
    in_mat = orbit.reshape(6, 1) + dmat

    refs = get_uint32_index(ring, refpts)
    out_mat = numpy.rollaxis(
        numpy.squeeze(internal_lpass(ring, in_mat, refpts=refs,
                                     keep_lattice=keep_lattice), axis=3), -1
    )
    # out_mat: 8 particles at n refpts for one turn
    # (x + d) - (x - d) / d
    m44 = (in_mat[:4, :4] - in_mat[:4, 4:]) / scaling

    if len(refs) > 0:
        mstack = (out_mat[:, :4, :4] - out_mat[:, :4, 4:]) / scaling
        if full:
            mstack = numpy.stack([mrotate(mat) for mat in mstack], axis=0)
    else:
        mstack = numpy.empty((0, 4, 4), dtype=float)

    return m44, mstack


@frequency_control
def find_m66(ring: Lattice, refpts: Refpts = None,
             orbit: Orbit = None, keep_lattice: bool = False, **kwargs):
    """One-turn 6x6 transfer matrix

    :py:func:`find_m66` finds the 6x6 transfer matrix of an accelerator
    lattice by differentiation of trajectories near the closed orbit.

    :py:func:`find_m66` uses :py:func:`.find_orbit6` to search for the closed
    orbit in 6-D. In order for this to work the ring **MUST** have a ``Cavity``
    element

    Parameters:
        ring:           Lattice description
        refpts:         Observation points
        orbit:          Avoids looking for initial the closed orbit if it is
          already known ((6,) array).
        keep_lattice:   Assume no lattice change since the previous tracking.
          Default: :py:obj:`False`

    Keyword Args:
        XYStep (float)  Step size.
          Default: :py:data:`DConstant.XYStep <.DConstant>`
        DPStep (float): Momentum step size.
          Default: :py:data:`DConstant.DPStep <.DConstant>`

    Returns:
        m66:    full one-turn matrix at the entrance of the first element
        ms:     6x6 transfer matrices between the entrance of the first
          element and each element indexed by refpts: (Nrefs, 6, 6) array

    See also:
         :py:func:`find_m44`, :py:func:`.find_orbit6`
    """
    xy_step = kwargs.pop('XYStep', DConstant.XYStep)
    dp_step = kwargs.pop('DPStep', DConstant.DPStep)
    if orbit is None:
        if ring.radiation:
            orbit, _ = find_orbit6(ring, keep_lattice=keep_lattice,
                                   XYStep=xy_step, DPStep=dp_step, **kwargs)
        else:
            orbit, _ = find_orbit4(ring, keep_lattice=keep_lattice,
                                   XYStep=xy_step, **kwargs)
        keep_lattice = True

    # Construct matrix of plus and minus deltas
    # scaling = 2*xy_step*numpy.array([1.0, 0.1, 1.0, 0.1, 1.0, 1.0])
    scaling = xy_step * numpy.array([1.0, 1.0, 1.0, 1.0, 0.0, 0.0]) + \
        dp_step * numpy.array([0.0, 0.0, 0.0, 0.0, 1.0, 1.0])
    dg = numpy.asfortranarray(0.5 * numpy.diag(scaling))
    dmat = numpy.concatenate((dg, -dg), axis=1)

    in_mat = orbit.reshape(6, 1) + dmat

    refs = get_uint32_index(ring, refpts)
    out_mat = numpy.rollaxis(
        numpy.squeeze(internal_lpass(ring, in_mat, refpts=refs,
                                     keep_lattice=keep_lattice), axis=3), -1
    )
    # out_mat: 12 particles at n refpts for one turn
    # (x + d) - (x - d) / d
    m66 = (in_mat[:, :6] - in_mat[:, 6:]) / scaling

    if len(refs) > 0:
        mstack = (out_mat[:, :, :6] - out_mat[:, :, 6:]) / scaling
    else:
        mstack = numpy.empty((0, 6, 6), dtype=float)

    return m66, mstack


def find_elem_m66(elem: Element, orbit: Orbit = None, **kwargs):
    """Single element 6x6 transfer matrix

    Numerically finds the 6x6 transfer matrix of a single element

    Parameters:
        elem:           AT element
        orbit:          Closed orbit at the entrance of the element,
          default: 0.0

    Keyword Args:
        XYStep (float): Step size.
          Default: :py:data:`DConstant.XYStep <.DConstant>`
        particle (Particle):    circulating particle.
          Default: :code:`lattice.particle` if existing,
          otherwise :code:`Particle('relativistic')`
        energy (float):         lattice energy. Default 0.

    Returns:
        m66:            6x6 transfer matrix
    """
    xy_step = kwargs.pop('XYStep', DConstant.XYStep)
    if orbit is None:
        orbit = numpy.zeros((6,))

    # Construct matrix of plus and minus deltas
    # scaling = 2*xy_step*numpy.array([1.0, 0.1, 1.0, 0.1, 1.0, 1.0])
    scaling = xy_step * numpy.array([1.0, 1.0, 1.0, 1.0, 1.0, 1.0])
    dg = numpy.asfortranarray(0.5 * numpy.diag(scaling))
    dmat = numpy.concatenate((dg, -dg), axis=1)

    in_mat = orbit.reshape(6, 1) + dmat
    internal_epass(elem, in_mat, **kwargs)
    m66 = (in_mat[:, :6] - in_mat[:, 6:]) / scaling
    return m66


def gen_m66_elem(ring: Lattice,
                 o4b: Orbit,
                 o4e: Orbit) -> M66:
    """Converts a ring to a linear 6x6 matrix

    Parameters:
        ring:       Lattice description
        o4b:
        o4e:

    Returns:
        m66:        6x6 transfer matrix
    """

    dipoles = ring[Dipole]
    theta = numpy.array([elem.BendingAngle for elem in dipoles])
    lendp = numpy.array([elem.Length for elem in dipoles])
    s_pos = ring.get_s_pos()
    s = numpy.diff(numpy.array([s_pos[0], s_pos[-1]]))[0]
    i2 = numpy.sum(numpy.abs(theta * theta / lendp))
    m66_mat, _ = find_m66(ring, [], orbit=o4b)
    if ring.radiation is False:
        m66_mat = symplectify(m66_mat)  # remove for damping
    lin_elem = M66('Linear', m66_mat, T1=-o4b, T2=o4e, Length=s, I2=i2)
    return lin_elem


Lattice.find_m44 = find_m44
Lattice.find_m66 = find_m66



from .atpass import atpass as _atpass, elempass as _elempass
from .utils import fortran_align
from .track import internal_lpass, internal_epass, internal_plpass
from .utils import initialize_lpass
from ..lattice import Lattice, Element, Particle, Refpts, End
from typing import Iterable, Optional
from warnings import warn


__all__ = ['element_pass', 'lattice_pass', 'patpass', 'atpass', 'elempass']


# noinspection PyIncorrectDocstring
def atpass(*args, **kwargs):
    """
    atpass(line, r_in, nturns, refpts=[], reuse=False, omp_num_threads=0)

    Track input particles *r_in* along line for *nturns* turns.

    Parameters:
        line (Sequence[Element]): list of elements
        r_in:                   6 x n_particles Fortran-ordered numpy array.
          On return, rin contains the final coordinates of the particles
        nturns (int):           number of turns to be tracked

    Keyword arguments:
        refpts (Uint32_refs):   numpy array of indices of elements where
          output is desired:

          * 0 means entrance of the first element
          * len(line) means end of the last element

        energy:                 nominal energy [eV]
        rest_energy:            rest_energy of the particle [eV]
        charge:                 particle charge [elementary charge]
        reuse (bool):           if True, use previously cached description
          of the lattice.
        omp_num_threads (int):  number of OpenMP threads
          (default 0: automatic)
        losses (bool):          if True, process losses

    Returns:
        r_out:  6 x n_particles x n_refpts x n_turns Fortran-ordered
          numpy array of particle coordinates

    :meta private:
    """
    warn(UserWarning("The public interface for tracking is 'lattice_pass'"))
    return _atpass(*args, **kwargs)


# noinspection PyIncorrectDocstring
def elempass(*args, **kwargs):
    """elempass(element, r_in)

    Track input particles *r_in* through a single element.

    Parameters:
        element (Element):  AT element
        rin:                6 x n_particles Fortran-ordered numpy array.
          On return, rin contains the final coordinates of the particles

    Keyword arguments:
        energy:             nominal energy [eV]
        rest_energy:        rest_energy of the particle [eV]
        charge:             particle charge [elementary charge]

    :meta private:
    """
    warn(UserWarning("The public interface for tracking is 'element_pass'"))
    return _elempass(*args, **kwargs)


def lattice_pass(lattice: Iterable[Element], r_in, nturns: int = 1,
                 refpts: Refpts = End, **kwargs):
    """
    :py:func:`lattice_pass` tracks particles through each element of a lattice
    calling the element-specific tracking function specified in the Element's
    *PassMethod* field.

    Parameters:
        lattice:                list of elements
        r_in:                   (6, N) array: input coordinates of N particles.
          *r_in* is modified in-place and reports the coordinates at
          the end of the element. For the best efficiency, *r_in*
          should be given as F_CONTIGUOUS numpy array.
        nturns:                 number of turns to be tracked
        refpts:                 Selects the location of coordinates output.
          See ":ref:`Selecting elements in a lattice <refpts>`"

    Keyword arguments:
        keep_lattice (bool):    Use elements persisted from a previous
          call. If :py:obj:`True`, assume that the lattice has not changed
          since the previous call.
        keep_counter (bool):    Keep the turn number from the previous
          call.
        turn (int):             Starting turn number. Ignored if
          *keep_counter* is :py:obj:`True`. The turn number is necessary to
          compute the absolute path length used in RFCavityPass.
        losses (bool):          Boolean to activate loss maps output
        omp_num_threads (int):  Number of OpenMP threads
          (default: automatic)

    The following keyword arguments overload the Lattice values

    Keyword arguments:

        particle (Optional[Particle]):  circulating particle.
          Default: :code:`lattice.particle` if existing,
          otherwise :code:`Particle('relativistic')`
        energy (Optiona[float]):        lattice energy. Default 0.
        unfold_beam (bool): Internal beam folding activate, this
            assumes the input particles are in bucket 0, works only
            if all bucket see the same RF Voltage.
            Default: :py:obj:`True`

    If *energy* is not available, relativistic tracking if forced,
    *rest_energy* is ignored.

    Returns:
        r_out: (6, N, R, T) array containing output coordinates of N particles
          at R reference points for T turns.
        loss_map: If *losses* is :py:obj:`True`: dictionary with the
          following key:

          ==============    ===================================================
          **islost**        (npart,) bool array indicating lost particles
          **turn**          (npart,) int array indicating the turn at
                            which the particle is lost
          **element**       ((npart,) int array indicating the element at
                            which the particle is lost
          **coord**         (6, npart) float array giving the coordinates at
                            which the particle is lost (zero for surviving
                            particles)
          ==============    ===================================================

    .. note::

       * :pycode:`lattice_pass(lattice, r_in, refpts=len(line))` is the same as
         :pycode:`lattice_pass(lattice, r_in)` since the reference point
         len(line) is the exit of the last element.
       * :pycode:`lattice_pass(lattice, r_in, refpts=0)` is a copy of *r_in*
         since the reference point 0 is the entrance of the first element.
       * To resume an interrupted tracking (for instance to get intermediate
         results), one must use one of the *turn* or *keep_counter*
         keywords to ensure the continuity of the turn number.
       * For multiparticle tracking with large number of turn the size of
         *r_out* may increase excessively. To avoid memory issues
         :pycode:`lattice_pass(lattice, r_in, refpts=None)` can be used.
         An empty list is returned and the tracking results of the last turn
         are stored in *r_in*.
       * To model buckets with different RF voltage :pycode:`unfold_beam=False`
         has to be used. The beam can be unfolded using the function
         :py:func:`.unfold_beam`. This function takes into account
         the true voltage in each bucket and distributes the particles in the
         bunches defined by :code:`ring.fillpattern` using a 6D orbit search.
    """
    lattice = initialize_lpass(lattice, nturns, kwargs)
    return internal_lpass(lattice, r_in, nturns=nturns, refpts=refpts,
                          no_varelem=False, **kwargs)


@fortran_align
def element_pass(element: Element, r_in, **kwargs):
    """Tracks particles through a single element.

    Parameters:
        element:                AT element
        r_in:                   (6, N) array: input coordinates of N particles.
          *r_in* is modified in-place and reports the coordinates at
          the end of the element. For the best efficiency, *r_in*
          should be given as F_CONTIGUOUS numpy array.

    Keyword arguments:
        particle (Particle):    circulating particle.
          Default: :code:`lattice.particle` if existing,
          otherwise :code:`Particle('relativistic')`
        energy (float):         lattice energy. Default 0.

    If *energy* is not available, relativistic tracking if forced,
    *rest_energy* is ignored.

    Returns:
        r_out:              (6, N) array containing output the coordinates of
          the particles at the exit of the element.
    """
    return internal_epass(element, r_in, **kwargs)


def patpass(lattice: Iterable[Element], r_in, nturns: int = 1,
            refpts: Refpts = End, pool_size: int = None,
            start_method: str = None, **kwargs):
    """
    Simple parallel implementation of :py:func:`.lattice_pass`.
    If more than one particle is supplied, use multiprocessing. For a
    single particle or if the lattice contains :py:class:`.Collective`
    elements, :py:func:`.atpass` is used.

    :py:func:`patpass` tracks particles through each element of a lattice
    calling the element-specific tracking function specified in the Element's
    *PassMethod* field.

    Parameters:
        lattice:                list of elements
        r_in:                   (6, N) array: input coordinates of N particles.
          *r_in* is modified in-place and reports the coordinates at
          the end of the element. For the best efficiency, *r_in*
          should be given as F_CONTIGUOUS numpy array.
        nturns:                 number of turns to be tracked
        refpts:                 Selects the location of coordinates output.
          See ":ref:`Selecting elements in a lattice <refpts>`"
        pool_size:              number of processes. If None,
          ``min(npart,nproc)`` is used
        start_method:           python multiprocessing start method.
          :py:obj:`None` uses the python default that is considered safe.
          Available values: ``'fork'``, ``'spawn'``, ``'forkserver'``.
          Default for linux is ``'fork'``, default for macOS and  Windows is
          ``'spawn'``. ``'fork'`` may be used on macOS to speed up the
          calculation or to solve Runtime Errors, however it is considered
          unsafe.

    Keyword arguments:
        keep_lattice (bool):    Use elements persisted from a previous
          call. If :py:obj:`True`, assume that the lattice has not changed
          since the previous call.
        keep_counter (bool):    Keep the turn number from the previous
          call.
        turn (int):             Starting turn number. Ignored if
          *keep_counter* is :py:obj:`True`. The turn number is necessary to
          compute the absolute path length used in RFCavityPass.
        losses (bool):          Boolean to activate loss maps output
        omp_num_threads (int):  Number of OpenMP threads
          (default: automatic)

    The following keyword arguments overload the Lattice values

    Keyword arguments:
        particle (Particle):    circulating particle.
          Default: *lattice.particle* if existing,
          otherwise *Particle('relativistic')*
        energy (float):         lattice energy. Default 0.

    If *energy* is not available, relativistic tracking if forced,
    *rest_energy* is ignored.

    Returns:
        r_out: (6, N, R, T) array containing output coordinates of N particles
          at R reference points for T turns.
        loss_map: If *losses* is :py:obj:`True`: dictionary with the
          following key:

          ==============    ===================================================
          **islost**        (npart,) bool array indicating lost particles
          **turn**          (npart,) int array indicating the turn at
                            which the particle is lost
          **element**       ((npart,) int array indicating the element at
                            which the particle is lost
          **coord**         (6, npart) float array giving the coordinates at
                            which the particle is lost (zero for surviving
                            particles)
          ==============    ===================================================

    .. note::

       * For multiparticle tracking with large number of turn the size of
         *r_out* may increase excessively. To avoid memory issues
         :pycode:`lattice_pass(lattice, r_in, refpts=[])` can be used.
         An empty list is returned and the tracking results of the last turn
         are stored in *r_in*.
       * By default, :py:func:`patpass` will use all the available CPUs.
         To change the number of cores used in ALL functions using
         :py:func:`patpass` (:py:mod:`~at.acceptance.acceptance` module for
         example) it is possible to set ``at.DConstant.patpass_poolsize``
         to the desired value.

    """
    kwargs['pool_size'] = pool_size
    kwargs['start_method'] = start_method
    lattice = initialize_lpass(lattice, nturns, kwargs)
    return internal_plpass(lattice, r_in, nturns=nturns,
                           refpts=refpts, **kwargs)


Lattice.lattice_pass = lattice_pass



"""
Conversion utilities for creating pyat elements
"""

from __future__ import annotations

__all__ = [
    "element_classes",
    "element_from_dict",
    "element_to_dict",
    "find_class",
    "keep_elements",
    "keep_attributes",
    "split_ignoring_parentheses",
    "RingParam",
    "protect",
    "restore",
]

import collections
import os
import re
import sysconfig
from typing import Any, Optional
from warnings import warn
from collections.abc import Callable, Generator

import numpy as np

from at import integrators
from at.lattice import AtWarning
from at.lattice import elements as elt
from at.lattice import Lattice, Particle, Element, Marker
from at.lattice import idtable_element

_ext_suffix = sysconfig.get_config_var("EXT_SUFFIX")
_placeholder = "placeholder"


def _no_encoder(v):
    """type encoding for .mat files"""
    return v


def _particle(value) -> Particle:
    if isinstance(value, Particle):
        # Create from python: save_mat
        return value
    elif isinstance(value, dict):
        # Create from Matlab: load_mat
        return Particle(**value)
    else:
        return Particle(value)


def _warn(index: int, message: str, elem_dict: dict) -> None:
    name = elem_dict.get("FamName", "")
    location = f'"{name}":\n' if index is None else f'{index} ("{name}"):\n'
    warning = "".join(("In element ", location, message, f"\n{elem_dict}\n"))
    warn(AtWarning(warning), stacklevel=2)


def element_classes() -> frozenset[type[Element]]:
    """Build a set of all Element subclasses"""

    # Misses class aliases (Bend, Matrix66)
    def subclasses_recursive(cl):
        direct = cl.__subclasses__()
        indirect = []
        for subclass in direct:
            indirect.extend(subclasses_recursive(subclass))
        return frozenset([cl] + direct + indirect)

    return subclasses_recursive(Element)


class RingParam(elt.Element):
    """Private class for Matlab RingParam element

    :meta private:
    """

    # noinspection PyProtectedMember
    _BUILD_ATTRIBUTES = elt.Element._BUILD_ATTRIBUTES + [
        "Energy",
        "Periodicity",
    ]
    _conversions = dict(
        elt.Element._conversions, Energy=float, Periodicity=int, Particle=_particle
    )

    # noinspection PyPep8Naming
    def __init__(
        self,
        FamName: str,
        Energy: float,
        Periodicity: int,
        **kwargs,
    ):
        if not np.isnan(float(Energy)):
            kwargs["Energy"] = Energy
        kwargs.setdefault("PassMethod", "IdentityPass")
        super(RingParam, self).__init__(FamName, Periodicity=Periodicity, **kwargs)


_alias_map = {
    "bend": elt.Dipole,
    "rbend": elt.Dipole,
    "sbend": elt.Dipole,
    "quad": elt.Quadrupole,
    "sext": elt.Sextupole,
    "rf": elt.RFCavity,
    "bpm": elt.Monitor,
    "ap": elt.Aperture,
    "ringparam": RingParam,
    "wig": elt.Wiggler,
    "matrix66": elt.M66,
    "M66": elt.M66,
}

# Map class names to Element classes
_CLASS_MAP = dict((cls.__name__.lower(), cls) for cls in element_classes())
_CLASS_MAP.update(_alias_map)

# Maps passmethods to Element classes
_PASS_MAP = {
    "BendLinearPass": elt.Dipole,
    "BndMPoleSymplectic4RadPass": elt.Dipole,
    "BndMPoleSymplectic4Pass": elt.Dipole,
    "QuadLinearPass": elt.Quadrupole,
    "StrMPoleSymplectic4Pass": elt.Multipole,
    "StrMPoleSymplectic4RadPass": elt.Multipole,
    "CorrectorPass": elt.Corrector,
    "CavityPass": elt.RFCavity,
    "RFCavityPass": elt.RFCavity,
    "ThinMPolePass": elt.ThinMultipole,
    "Matrix66Pass": elt.M66,
    "AperturePass": elt.Aperture,
    "IdTablePass": idtable_element.InsertionDeviceKickMap,
    "GWigSymplecticPass": elt.Wiggler,
}

# Maps python class name to Matlab class
# Default: element_class.__name__
_mat_class = {
    "Dipole": "Bend",
    "M66": "Matrix66",
}

# Lattice attributes which must be dropped when writing a file
_drop_attrs = {
    "in_file": None,
    "use": None,
    "mat_key": None,
    "mat_file": None,  # Not used anymore...
    "m_file": None,
    "repr_file": None,
}


def _hasattrs(kwargs: dict, *attributes) -> bool:
    """Checks the presence of keys in a :py:class:`dict`

    Returns :py:obj:`True` if any of the ``attributes`` is in ``kwargs``

    Args:
        kwargs:     The dictionary of keyword arguments passed to the
          Element constructor.
        attributes: A list of strings, the attribute names to be checked.

    Returns:
        found (bool):   :py:obj:`True` if the element has any of the specified
          attributes.
    """
    for attribute in attributes:
        if attribute in kwargs:
            return True
    return False


def keep_attributes(ring: Lattice):
    """Remove Lattice attributes which must not be saved on file"""
    return dict(
        (k, v) for k, v in ring.attrs.items() if _drop_attrs.get(k, k) is not None
    )


def keep_elements(ring: Lattice) -> Generator[Element, None, None]:
    """Remove the 'RingParam' Marker"""
    for elem in ring:
        if not (isinstance(elem, Marker) and getattr(elem, "tag", None) == "RingParam"):
            yield elem


def _from_contents(elem: dict) -> type[Element]:
    """Deduce the element class from its contents"""

    def low_order(key):
        polynom = np.array(elem[key], dtype=np.float64).reshape(-1)
        try:
            low = np.where(polynom != 0.0)[0][0]
        except IndexError:
            low = -1
        return low

    length = float(elem.get("Length", 0.0))
    pass_method = elem.get("PassMethod", "")
    if _hasattrs(
        elem, "FullGap", "FringeInt1", "FringeInt2", "gK", "EntranceAngle", "ExitAngle"
    ):
        return elt.Dipole
    elif _hasattrs(elem, "Voltage", "Frequency", "HarmNumber", "PhaseLag", "TimeLag"):
        return elt.RFCavity
    elif _hasattrs(elem, "Periodicity"):
        # noinspection PyProtectedMember
        return RingParam
    elif _hasattrs(elem, "Limits"):
        return elt.Aperture
    elif _hasattrs(elem, "M66"):
        return elt.M66
    elif _hasattrs(elem, "K"):
        return elt.Quadrupole
    elif _hasattrs(elem, "PolynomB", "PolynomA"):
        loworder = low_order("PolynomB")
        if loworder == 1:
            return elt.Quadrupole
        elif loworder == 2:
            return elt.Sextupole
        elif loworder == 3:
            return elt.Octupole
        elif pass_method.startswith("StrMPoleSymplectic4") or (length > 0):
            return elt.Multipole
        else:
            return elt.ThinMultipole
    elif _hasattrs(elem, "KickAngle"):
        return elt.Corrector
    elif length > 0.0:
        return elt.Drift
    elif _hasattrs(elem, "GCR"):
        return elt.Monitor
    elif pass_method == "IdentityPass":
        return elt.Marker
    else:
        return elt.Element


def find_class(
    elem_dict: dict, quiet: bool = False, index: Optional[int] = None
) -> type(Element):
    """Deduce the class of an element from its attributes

    `find_class` looks first at the "Class" field, if existing. It then tries to deduce
    the class from "FamName", from "PassMethod", and finally form the element contents.

    Args:
        elem_dict:      The dictionary of keyword arguments passed to the
                        Element constructor.
        quiet:          Suppress the warning for non-standard classes
        index:          Element index in the lattice

    Returns:
        element_class:  The guessed Class name
    """

    def check_class(clname):
        if clname:
            _warn(index, f"Class '{clname}' does not exist.", elem_dict)

    def check_pass(passm):
        if not passm:
            _warn(index, "No PassMethod provided.", elem_dict)
        elif not passm.endswith("Pass"):
            message = (
                f"Invalid PassMethod '{passm}': "
                "provided pass methods should end in 'Pass'."
            )
            _warn(index, message, elem_dict)

    class_name = elem_dict.pop("Class", "")  # try from class name
    cls = _CLASS_MAP.get(class_name.lower(), None)
    if cls is not None:
        return cls
    elif not quiet:
        check_class(class_name)

    elname = elem_dict.get("FamName", "")  # try from element name
    cls = _CLASS_MAP.get(elname.lower(), None)
    if cls is not None:
        return cls

    pass_method = elem_dict.get("PassMethod", "")  # try from passmethod
    cls = _PASS_MAP.get(pass_method, None)
    if cls is not None:
        return cls
    elif not quiet:
        check_pass(pass_method)

    return _from_contents(elem_dict)  # look for contents


def element_from_dict(
    elem_dict: dict,
    index: Optional[int] = None,
    check: bool = True,
    quiet: bool = False,
) -> Element:
    """Builds an :py:class:`.Element` from a dictionary of attributes

    Parameters:
        elem_dict:      Dictionary of element attributes
        index:          Element index
        check:          Check the compatibility of class and PassMethod
        quiet:          Suppress the warning for non-standard classes

    Returns:
        elem (Element): new :py:class:`.Element`
    """

    # noinspection PyShadowingNames
    def sanitise_class(index, cls, elem_dict):
        """Checks that the Class and PassMethod of the element are a valid
            combination. Some Classes and PassMethods are incompatible and
            would raise errors during calculation, so we send a
            warning here.

        Args:
            index:          element index
            cls:            Proposed class
            elem_dict:      The dictionary of keyword arguments passed to the
                            Element constructor.

        Raises:
            AttributeError: if the PassMethod and Class are incompatible.
        """
        class_name = cls.__name__
        pass_method = elem_dict.get("PassMethod")
        if pass_method is not None:
            pass_to_class = _PASS_MAP.get(pass_method)
            length = float(elem_dict.get("Length", 0.0))
            file_name = pass_method + _ext_suffix
            file_path = os.path.join(integrators.__path__[0], file_name)
            if not os.path.isfile(os.path.realpath(file_path)):
                message = f"PassMethod {pass_method} is missing {file_name}."
                _warn(index, message, elem_dict)
            elif (pass_method == "IdentityPass") and (length != 0.0):
                message = (
                    f"PassMethod {pass_method} is not compatible with length {length}."
                )
                _warn(index, message, elem_dict)
            elif pass_to_class is not None:
                if not issubclass(cls, pass_to_class):
                    message = (
                        f"PassMethod {pass_method} is not compatible "
                        f"with Class {class_name}."
                    )
                    _warn(index, message, elem_dict)

    cls = find_class(elem_dict, quiet=quiet, index=index)
    if check:
        sanitise_class(index, cls, elem_dict)
    # Remove mandatory attributes from the keyword arguments.
    # Create list rather than generator to ensure that elements are removed
    # from elem_dict.
    elem_args = [elem_dict.pop(attr, None) for attr in cls._BUILD_ATTRIBUTES]
    element = cls(*(arg for arg in elem_args if arg is not None), **elem_dict)
    return element


def element_to_dict(elem: Element, encoder: Callable[[Any], Any] = _no_encoder) -> dict:
    """Convert a :py:class:`.Element` to a :py:class:`dict`

    Parameters:
        elem:           :py:class:`.Element`
        encoder:        data converter

    Returns:
        dct (dict):     Dictionary of :py:class:`.Element` attributes
    """
    dct = dict((k, encoder(v)) for k, v in elem.items())
    class_name = elem.__class__.__name__
    dct["Class"] = _mat_class.get(class_name, class_name)
    return dct


def split_ignoring_parentheses(
        string: str,
        delimiter: str = ",",
        fence: tuple[str, str] = ('\\(', '\\)'),
        maxsplit: int = -1,
) -> list[str]:
    """Split a string while keeping protected expressions intact

    Example: "l=0,hom(4,0.0,0)" -> ["l=0", "hom(4,0.0,0)"]
    """
    substituted, matches = protect(string, fence=fence)
    parts = substituted.split(delimiter, maxsplit=maxsplit)
    return restore(matches, *parts)


def protect(string: str, fence: tuple[str, str] = ('"', '"')):
    inf, outf = fence
    pattern = f"{inf}.*?{outf}"
    substituted = string[:]
    matches = collections.deque(re.finditer(pattern, string))
    for match in matches:
        substituted = substituted.replace(match.group(), _placeholder, 1)
    return substituted, matches


def restore(matches, *parts):

    def rep(part):
        while _placeholder in part:
            next_match = matches.popleft()
            part = part.replace(_placeholder, next_match.group(), 1)
        return part

    replaced_parts = [rep(part) for part in parts]
    assert not matches
    return replaced_parts


Element.from_dict = staticmethod(element_from_dict)
Element.to_dict = element_to_dict



"""Load a lattice from an Elegant file (.lte).

This is not complete but can parse the example files that I have.
This parser is quite similar to the Tracy parser in tracy.py.

The Elegant file format is described briefly `here
<https://ops.aps.anl.gov/manuals/elegant_latest/elegantse9.html#x113-1120009>`_.
It is similar to the `MAD-X format <http://madx.web.cern.ch/madx/>`_.

Note that Elegant scales magnet polynomials in a different way
to AT, so the parsed coefficients need to be divided by n! for
the coefficient of order n.

"""
import logging as log
from os.path import abspath
import re

from at.lattice.elements import (
    Aperture,
    Corrector,
    Drift,
    Dipole,
    Marker,
    Multipole,
    Octupole,
    Quadrupole,
    RFCavity,
    Sextupole,
)
from at.lattice import Lattice
from at.load import register_format, utils

__all__ = ['load_elegant']


# noinspection PyUnusedLocal
def create_drift(name, params, energy, harmonic_number):
    length = params.pop("l", 0)
    return Drift(name, length, **params)


# noinspection PyUnusedLocal
def create_marker(name, params, energy, harmonic_number):
    return Marker(name, **params)


# noinspection PyUnusedLocal
def create_aperture(name, params, energy, harmonic_number):
    x_lim = float(params.get('x_max'))
    y_lim = float(params.get('y_max'))
    limits = [-x_lim, x_lim, -y_lim, y_lim]
    return Aperture(name, limits)


# noinspection PyUnusedLocal
def create_quad(name, params, energy, harmonic_number):
    length = params.pop("l", 0)
    params["NumIntSteps"] = params.pop("n_kicks", 10)
    params["k"] = float(params.pop("k1"))
    params["PassMethod"] = "StrMPoleSymplectic4Pass"
    return Quadrupole(name, length, **params)


# noinspection PyUnusedLocal
def create_sext(name, params, energy, harmonic_number):
    length = params.pop("l", 0)
    params["NumIntSteps"] = params.pop("n_kicks", 10)
    k2 = float(params.pop("k2", 0))
    return Sextupole(name, length, k2 / 2, **params)


# noinspection PyUnusedLocal
def create_oct(name, params, energy, harmonic_number):
    length = params.pop("l", 0)
    params["NumIntSteps"] = params.pop("n_kicks", 10)
    k3 = float(params.pop("k3", 0))
    PolynomA = [0, 0, 0, 0]
    PolynomB = [0, 0, 0, k3 / 6]
    return Octupole(name, length, PolynomA, PolynomB, **params)


# noinspection PyUnusedLocal
def create_multipole(name, params, energy, harmonic_number):
    def factorial(x, acc=1):
        if x == 0:
            return acc
        else:
            return factorial(x - 1, acc * x)

    length = params.pop("l", 0)
    PolynomA = [0, 0, 0, 0]
    PolynomB = [0, 0, 0, 0]
    if "knl" in params:
        order = int(float(params.pop("order")))
        PolynomB[order] = float(params.pop("knl")) / factorial(order)
    if "ksl" in params:
        order = int(float(params.pop("order")))
        PolynomA[order] = float(params.pop("ksl")) / factorial(order)

    return Multipole(name, length, PolynomA, PolynomB, **params)


# noinspection PyUnusedLocal
def create_dipole(name, params, energy, harmonic_number):
    length = params.pop("l", 0)
    params["NumIntSteps"] = params.pop("n_kicks", 10)
    params["PassMethod"] = "BndMPoleSymplectic4Pass"
    params["BendingAngle"] = float(params.pop("angle"))
    params["EntranceAngle"] = float(params.pop("e1"))
    params["ExitAngle"] = float(params.pop("e2"))
    if "hgap" in params:
        params["FullGap"] = float(params.pop("hgap")) * 2
        # What should we do if no fint property?
        fint = params.pop("fint", 1)
        params["FringeInt1"] = fint
        params["FringeInt2"] = fint
    k1 = float(params.pop("k1", 0))
    k2 = float(params.pop("k2", 0))
    k3 = float(params.pop("k3", 0))
    k4 = float(params.pop("k4", 0))
    params["PolynomB"] = [0, k1, k2 / 2, k3 / 6, k4 / 24]
    return Dipole(name, length, **params)


# noinspection PyUnusedLocal
def create_corrector(name, params, energy, harmonic_number):
    length = params.pop("l", 0)
    hkick = params.pop("hkick", 0)
    vkick = params.pop("vkick", 0)
    kick_angle = [hkick, vkick]
    return Corrector(name, length, kick_angle, **params)


def create_cavity(name, params, energy, harmonic_number):
    length = params.pop("l", 0)
    voltage = params.pop("volt")
    frequency = params.pop("freq")
    params["Phi"] = params.pop("phase")
    return RFCavity(name, length, voltage,
                    frequency, harmonic_number, energy, **params)


ELEMENT_MAP = {
    "drift": create_drift,
    "drif": create_drift,
    "edrift": create_drift,
    # This should be a wiggler.
    "cwiggler": create_drift,
    "csben": create_dipole,
    "csbend": create_dipole,
    "csrcsben": create_dipole,
    "quadrupole": create_quad,
    "kquad": create_quad,
    "ksext": create_sext,
    "koct": create_oct,
    "mult": create_multipole,
    "multipole": create_multipole,
    "kicker": create_corrector,
    "hkick": create_corrector,
    "vkick": create_corrector,
    "mark": create_marker,
    "marker": create_marker,
    "malign": create_marker,
    "recirc": create_marker,
    "sreffects": create_marker,
    "rcol": create_marker,
    "watch": create_marker,
    "charge": create_marker,
    "monitor": create_marker,
    "moni": create_marker,
    "maxamp": create_aperture,
    "rfca": create_cavity,
}


def parse_lines(contents):
    """Return individual lines.

    Remove comments and whitespace and convert to lowercase.
    """
    lines = []
    current_line = ""
    for line in contents.splitlines():
        line = line.lower()
        line = line.split("!")[0]
        line = line.strip()
        if line:
            if line.endswith("&"):
                current_line += line[:-1]
            else:
                lines.append(current_line + line)
                current_line = ""

    return lines


def parse_chunk(value, elements, chunks):
    """Parse a non-element part of a lattice file.

    That part can reference already-parsed elements and chunks.

    if
    chunks = {"x": ["a", "b"]}
    and
    elements = {"a": Marker(...), "b": Quadrupole(...)}

    line(a,b) => [a, b]
    -x        => [b, a]
    2*x       => [a, b, a, b]

    """
    chunk = []
    parts = utils.split_ignoring_parentheses(value, ",")
    for part in parts:
        # What's this?
        if "symmetry" in part:
            continue
        if "line" in part:
            line_parts = re.match("line\\s*=\\s*\\((.*)\\)", part).groups()[0]
            for p in line_parts.split(","):
                p = p.strip()
                chunk.extend(parse_chunk(p, elements, chunks))
        elif part.startswith("-"):
            # Reverse a sequence of elements. When doing this, swap
            # the entrance and exit angles for any dipoles.
            chunk_name = part[1:]
            try:
                chunk_to_invert = chunks[chunk_name]
            except KeyError:
                # You can reverse just a single element: probably a
                # bend, as there's no reason to invert other elements.
                chunk_to_invert = [elements[chunk_name]]
            inverted_chunk = []
            for el in reversed(chunk_to_invert):
                if el.__class__ == Dipole:
                    inverted_dipole = el.copy()
                    inverted_dipole.EntranceAngle = el.ExitAngle
                    inverted_dipole.ExitAngle = el.EntranceAngle
                    inverted_chunk.append(inverted_dipole)
                else:
                    inverted_chunk.append(el)
            chunk.extend(inverted_chunk)
        elif "*" in part:
            num, chunk_name = part.split("*")
            if chunk_name[0] == "(":
                assert chunk_name[-1] == ")"
                chunk_name = chunk_name[1:-1]
            chunk.extend(int(num) * parse_chunk(chunk_name, elements, chunks))
        elif part in elements:
            chunk.append(elements[part])
        elif part in chunks:
            chunk.extend(chunks[part])
        else:
            raise Exception("part {} not understood".format(part))
    return chunk


def expand_elegant(contents, lattice_key, energy, harmonic_number):
    lines = parse_lines(contents)
    variables = {"energy": energy, "harmonic_number": harmonic_number}
    elements = {}
    chunks = {}
    for line in lines:
        if ":" not in line:
            key, value = line.split("=")
            variables[key.strip()] = value.strip()
        else:
            key, value = line.split(":")
            key = key.strip()
            value = value.strip()
            if value.split(",")[0] in ELEMENT_MAP:
                elements[key] = elegant_element_from_string(key,
                                                            value,
                                                            variables)
            else:
                chunk = parse_chunk(value, elements, chunks)
                chunks[key] = chunk

    return chunks[lattice_key.lower()]


def handle_value(value):
    value = value.strip()
    if value.startswith('"'):
        assert value.endswith('"')
        value = value[1:-1]
        value = value.split()
        if len(value) > 1:
            # Handle basic arithmetic e.g. "0.04 2 /" -> 0.02
            assert len(value) == 3
            if value[2] == "/":
                value = float(value[0]) / float(value[1])
            elif value[2] == "*":
                value = float(value[0]) * float(value[1])
            elif value[2] == "-":
                value = float(value[0]) - float(value[1])
            elif value[2] == "+":
                value = float(value[0]) + float(value[1])
        else:
            value = value[0]
    return value


def elegant_element_from_string(name, element_string, variables):
    """Create element from Elegant's string representation.

    e.g. drift,l=0.045 => Drift(name, 0.045)

    """
    log.debug("Parsing elegant element {}".format(element_string))
    parts = utils.split_ignoring_parentheses(element_string, ",")
    params = {}
    element_type = parts[0]
    for part in parts[1:]:
        key, value = utils.split_ignoring_parentheses(part, "=")
        key = key.strip()
        value = handle_value(value)
        if value in variables:
            value = variables[value]

        params[key] = value

    energy = variables["energy"]
    harmonic_number = variables["harmonic_number"]

    return ELEMENT_MAP[element_type](name, params, energy, harmonic_number)


def load_elegant(filename: str, **kwargs) -> Lattice:
    """Create a :py:class:`.Lattice`  from an Elegant file

    Parameters:
        filename:           Name of an Elegant file

    Keyword Args:
        name (str):         Name of the lattice. Default: taken from
          the file.
        energy (float):     Energy of the lattice [eV]
        periodicity(int):   Number of periods. Default: taken from the elements, or 1
        *:                  All other keywords will be set as Lattice attributes

    Returns:
        lattice (Lattice):  New :py:class:`.Lattice` object

    See Also:
        :py:func:`.load_lattice` for a generic lattice-loading function.
    """
    try:
        energy = kwargs.pop("energy")
        lattice_key = kwargs.pop("lattice_key")
        harmonic_number = kwargs.pop("harmonic_number")

        def elem_iterator(params, elegant_file):
            with open(params.setdefault("in_file", elegant_file)) as f:
                contents = f.read()
                element_lines = expand_elegant(
                    contents, lattice_key, energy, harmonic_number
                )
                params.setdefault("energy", energy)
                for line in element_lines:
                    yield line

        lat = Lattice(abspath(filename), iterator=elem_iterator, **kwargs)
        return lat
    except Exception as e:
        raise ValueError('Failed to load elegant '
                         'lattice {}: {}'.format(filename, e))


register_format(
    ".lte", load_elegant, descr="Elegant format. See :py:func:`.load_elegant`.")



import numpy
import at
import matplotlib.pyplot as plt
from at.collective import Wake, LongResonatorElement
from at.collective.haissinski import Haissinski

# First we define the ring, the BB resonator, the current and the wake element
ring = at.load_m('../../../machine_data/esrf.m')



freq = 10e9
qfactor = 1
Rs = 1e4
current = 5e-4
m = 60  # 30 is quite coarse, 70 or 80 is very fine. 50 is middle
kmax = 8


srange = Wake.build_srange(0., 0.36, 1.0e-5, 1.0e-2,
                           ring.circumference, ring.circumference)

# Now we initialise the Haissinski class, and solve
# then we normalise the distribution and shift the charge center to be at 0
wobj = Wake.long_resonator(srange, freq, qfactor, Rs, ring.beta)
ha = Haissinski(wobj, ring, m=m, kmax=kmax,
                current=current, numIters=30, eps=1e-13)
ha.solve()

ha_x_tmp = ha.q_array*ha.sigma_l
ha_prof = ha.res/ha.Ic
ha_prof /= numpy.trapz(ha_prof, x=ha_x_tmp)
ha_cc = numpy.average(ha_x_tmp, weights=ha_prof)
ha_x = (ha_x_tmp - ha_cc)

'''
currents = numpy.arange(0, 1.1e-3, 2e-4)
ha.solve_steps(currents)

fig = plt.figure()
ax1 = fig.add_subplot(111)
for i in numpy.arange(len(currents)):
    ax1.plot(1e3*ha.q_array*ha.sigma_l, ha.res_steps[i,:]/ha.I_steps[i],
             label='Ib={:f}mA'.format(currents[i]*1e3),
             color=cm.jet(float(i)/len(currents)))
ax1.legend()
ax1.set_xlabel('z [mm]')
ax1.set_ylabel(r'$\rho(z)$')
plt.show()
'''


# Now we set up and run the tracking.
# The final distribution is an average of the last numAve turns
welem = LongResonatorElement('wake', ring,
                             srange, freq, qfactor, Rs, Nslice=300)


Nbunches = 1
ring.beam_current = current
ring.set_fillpattern(Nbunches)

ring.enable_6d()
ring.set_cavity_phase()
_, fring = at.fast_ring(ring)
fring.append(welem)


sigm = at.sigma_matrix(ring.radiation_on(copy=True))
Nparts = 10000
Nturns = 20000
nbins = 60
part = at.beam(Nparts, sigm)

histAveFlag = False
numAve = 5000
id0 = 0
for t in numpy.arange(Nturns):
    if t % 1000 == 0:
        print('Tracking turn ', t, ' of ', Nturns)
    fring.track(part, in_place=True, refpts=None)

    if t > Nturns-numAve:

        if not histAveFlag:
            histAveFlag = True

            nt1, et1 = numpy.histogram(part[5, :], bins=nbins)

            etc1 = numpy.array([(et1[i] + et1[i + 1]) / 2
                                for i in numpy.arange(len(et1) - 1)])

            allData = numpy.zeros((len(etc1), numAve + 1))
            allData[:, 0] = etc1
            allData[:, id0 + 1] = nt1

        else:
            n1, e1 = numpy.histogram(part[5, :], bins=et1)
            ec1 = numpy.array([(e1[i] + e1[i + 1]) / 2
                               for i in numpy.arange(len(e1) - 1)])
            allData[:, id0 + 1] = n1
        id0 += 1


# Identical post processing of tracking distribution (to the Haissinski)
zr_tmp = allData[:, 0]
prof = numpy.mean(allData[:, 1:], axis=1)
prof /= numpy.trapz(prof, x=zr_tmp)

cc = numpy.average(zr_tmp, weights=prof)
zr = zr_tmp - cc

fig = plt.figure()
ax1 = fig.add_subplot(111)
ax1.plot(1e3*ha_x, ha_prof,
         color='r', linestyle='solid', label='Haissinski Solution')
ax1.plot(1e3*zr, prof, color='k', linestyle='dashed', label='Tracking')
ax1.set_xlabel('z [mm]')
ax1.set_ylabel(r'$\rho(z)$')
ax1.legend()
plt.savefig('./haissinski_dist.png')
plt.show()



"""Global set of constants"""
from ..cconfig import ismpi, isopenmp
from numpy.random import Generator, PCG64, SeedSequence
import os
import multiprocessing

if ismpi():
    from mpi4py import MPI
    _comm = MPI.COMM_WORLD
    _MPI_sz = _comm.Get_size()
    _MPI_rk = _comm.Get_rank()
else:
    _MPI_sz = 1
    _MPI_rk = 0


def _newgen(seed=12345):
    ss = SeedSequence(seed)
    seeds = ss.spawn(_MPI_sz + 1)
    return Generator(PCG64(seeds[0])), Generator(PCG64(seeds[_MPI_rk + 1]))


class _Dst(object):
    """Set of constants for AT numerical analysis"""
    # optimized values:
    # see https://github.com/atcollab/at/pull/199#issuecomment-745420443
    XYStep = 3.e-8           # Coordinate step for differentiation
    DPStep = 3.e-6           # Momentum step for dispersion and chromaticity
    OrbConvergence = 1.e-12  # Convergence criterion for orbit
    OrbMaxIter = 20          # Max. number of iterations for orbit
    omp_num_threads = int(os.environ.get('OMP_NUM_THREADS', '0'))
    patpass_poolsize = multiprocessing.cpu_count()
    patpass_startmethod = None
    _rank = _MPI_rk         # MPI rank

    def __setattr__(self, name, value):
        _ = getattr(self, name)     # make sure attribute exists
        object.__setattr__(self, name, value)

    def reset(self, name):
        delattr(self, name)

    @property
    def mpi(self):
        return ismpi()

    @property
    def openmp(self):
        return isopenmp()

    @property
    def rank(self):
        return self._rank


class _Random(object):
    """Random generators for AT"""
    def __init__(self):
        self.reset()

    @classmethod
    def reset(cls, seed=12345):
        cls.common, cls.thread = _newgen(seed)


DConstant = _Dst()
"""
Default values for AT algorithms

Attributes:
    XYStep:              Coordinate step for differentiation
    DPStep:              Momentum step for dispersion and chromaticity
    OrbConvergence:      Convergence criterion for orbit
    OrbMaxIter:          Max. number of iterations for orbit
    omp_num_threads:     Default number of OpenMP threads
    patpass_poolsize:    Default size of multiprocessing pool
    patpass_startmethod: Default start method for the multiprocessing
    mpi:                 :py:obj:`True` if MPI is active
    openmp:              :py:obj:`True` if OpenMP is active

Methods:
    reset(attrname):    Reset the attribute to its default value
"""

random = _Random()
"""
Random generators for AT

Attributes:
    common(:py:class:`~numpy.random.Generator`): random generator common to
        all threads
    thread(:py:class:`~numpy.random.Generator`): random generator specific to
        each thread

Methods:
    reset(seed=None):   Reset the generators.

Examples:
    >>> at.random.common.random()
    0.8699988509120198

    Generates a random number in [0, 1) which is identical in all threads

    >>> at.random.thread.normal(size=5)
    array([-0.02326275, -0.89806482, -1.69086929, -0.74399398,  0.70156743])

    Generates an array of Gaussian random values which is specific to
    each thread
"""



import os
from tempfile import mktemp

import pytest
from numpy.testing import assert_equal

from at.lattice import Lattice


@pytest.mark.parametrize("lattice", ["dba_lattice", "hmba_lattice"])
@pytest.mark.parametrize(
    "suffix, options",
    [
        (".m", {}),
        (".repr", {}),
        (".mat", {"use": "abcd"}),
        (".json", {}),
    ],
)
def test_m(request, lattice, suffix, options):
    ring0 = request.getfixturevalue(lattice)
    fname = mktemp(suffix=suffix)

    # Create a new .m or .repr file
    ring0.save(fname, **options)

    # load the new file
    ring1 = Lattice.load(fname, **options)

    # Check that we get back the original lattice
    el1, rg1, _ = ring0.linopt6()
    el2, rg2, _ = ring1.linopt6()
    assert_equal(rg1.tune, rg2.tune)
    assert_equal(rg1.chromaticity, rg2.chromaticity)

    os.unlink(fname)



import numpy

'''
This is an example of a python passmethod, here the Identity
passmethod.

The python passmethods has to be in the python path or in
at.integratorsand contain a function trackFunction()

The element creation is as follows:

params = {'Length':0,
          'PassMethod':'pyIdentityPass',
          }

ib_elem = Element('py_id',**params)

To be noted:
The particles are described by a 2 dimensional vector
(6,num_particles)
In case a pyhton passmethod and a c passmethod with the same name
are found, the c passmethod is used
'''


_id = numpy.identity(6)


def trackFunction(rin, elem=None):
    for rtmp in numpy.atleast_2d(rin.T):
        rtmp[:] = numpy.dot(_id, rtmp)



"""
Non-linear optics
"""
import numpy
from typing import Optional, Sequence
from scipy.special import factorial
from ..lattice import Element, Lattice
from ..tracking import internal_lpass
from .orbit import Orbit, find_orbit
from .linear import get_tune, get_chrom, linopt6
from .harmonic_analysis import get_tunes_harmonic

__all__ = ['detuning', 'chromaticity', 'gen_detuning_elem', 'tunes_vs_amp']


def tunes_vs_amp(ring: Lattice, amp: Optional[Sequence[float]] = None,
                 dim: Optional[int] = 0, nturns: Optional[int] = 512,
                 **kwargs) -> numpy.ndarray:
    r"""Generates a range of tunes for varying x, y, or z amplitudes

    Parameters:
        ring:               Lattice description
        amp:                Oscillation amplitude. If amp is None the
                            tunes are calculated from :py:func:`.linopt6`
                            else tracking is used
        dim:                Coordinate index of the oscillation
        nturns:             Number of turns
        method:             ``'laskar'`` or ``'fft'``. Default: ``'laskar'``
        num_harmonics:      Number of harmonic components to compute (before
                            mask applied)
        fmin:               Lower bound for tune
        fmax:               Upper bound for tune
        hann:               Turn on Hanning window. Default: :py:obj:`False`

    Returns:
        tunes (ndarray):    Array of tunes qx, qy with shape (len(amp), 2)
    """

    def _gen_part(ring, amp, dim, orbit, ld, nturns):
        invx = numpy.array([[1/numpy.sqrt(ld['beta'][0]), 0],
                            [ld['alpha'][0]/numpy.sqrt(ld['beta'][0]),
                            numpy.sqrt(ld['beta'][0])]])

        invy = numpy.array([[1/numpy.sqrt(ld['beta'][1]), 0],
                            [ld['alpha'][1]/numpy.sqrt(ld['beta'][1]),
                            numpy.sqrt(ld['beta'][1])]])
        part = numpy.array([orbit, ] * len(amp)).T + 1.0e-6
        part[dim, :] += amp
        part = internal_lpass(ring, part, nturns=nturns)
        sh = part.shape
        partx = numpy.reshape(part[0, :], (sh[1], sh[3]))
        partxp = numpy.reshape(part[1, :], (sh[1], sh[3]))
        party = numpy.reshape(part[2, :], (sh[1], sh[3]))
        partyp = numpy.reshape(part[3, :], (sh[1], sh[3]))
        px = numpy.array([numpy.matmul(invx, [partx[i], partxp[i]])
                          for i in range(len(amp))])
        py = numpy.array([numpy.matmul(invy, [party[i], partyp[i]])
                          for i in range(len(amp))])
        return px[:, 0, :] - 1j*px[:, 1, :], py[:, 0, :] - 1j*py[:, 1, :]

    l0, bd, _ = linopt6(ring)
    orbit = l0['closed_orbit']
    tunes = bd['tune']

    if amp is not None:
        partx, party = _gen_part(ring, amp, dim, orbit, l0, nturns)
        qx = get_tunes_harmonic(partx, **kwargs)
        qy = get_tunes_harmonic(party, **kwargs)
        tunes = numpy.vstack((qx, qy)).T

    return numpy.array(tunes)


def detuning(ring: Lattice,
             xm: Optional[float] = 0.3e-4, ym: Optional[float] = 0.3e-4,
             npoints: Optional[int] = 3,
             nturns: Optional[int] = 512, **kwargs):
    """Computes the tunes for a sequence of amplitudes

    This function uses :py:func:`tunes_vs_amp` to compute the tunes for
    the specified amplitudes. Then it fits this data and returns
    the detuning coefficiant dQx/dx, dQy/dx, dQx/dy, dQy/dy and the
    qx, qy arrays versus x, y arrays

    Parameters:
        ring:       Lattice description
        xm:         Maximum x amplitude
        ym:         Maximum y amplitude
        npoints:    Number of points in each plane
        nturns:     Number of turns for tracking
        method:     ``'laskar'`` or ``'fft'``. Default: ``'laskar'``
        num_harmonics:  Number of harmonic components to compute
                       (before mask applied)
        fmin:       Lower bound for tune
        fmax:       Upper bound for tune
        hann:       Turn on Hanning window. Default: :py:obj:`False`

    Returns:
        q0 (ndarray): qx, qy from horizontal and vertical amplitude scans.
                      Fit results with shape (2, 2)
        q1 (ndarray): Amplitude detuning coefficients with shape (2, 2)
                      ``[[dQx/dx, dQy/dx], [dQx/dy, dQy/dy]]``
        x (ndarray): x amplitudes (npoints, )
        q_dx (ndarray): qx, qy tunes as a function of x amplitude (npoints, 2)
        y (ndarray): y amplitudes (npoints, )
        q_dy (ndarray): qx, qy tunes as a function of y amplitude (npoints, 2)
    """
    lindata0, _, _ = linopt6(ring)
    gamma = (1 + lindata0.alpha * lindata0.alpha) / lindata0.beta

    x = numpy.linspace(-xm, xm, npoints)
    y = numpy.linspace(-ym, ym, npoints)
    x2 = x * x
    y2 = y * y

    q_dx = tunes_vs_amp(ring, amp=x, dim=0, nturns=nturns, **kwargs)
    q_dy = tunes_vs_amp(ring, amp=y, dim=2, nturns=nturns, **kwargs)

    idx = numpy.isfinite(q_dx[:, 0]) & numpy.isfinite(q_dx[:, 1])
    idy = numpy.isfinite(q_dy[:, 0]) & numpy.isfinite(q_dy[:, 1])

    fx = numpy.polyfit(x2[idx], q_dx[idx], 1)
    fy = numpy.polyfit(y2[idy], q_dy[idy], 1)

    q0 = [[fx[1, 0], fx[1, 1]],
          [fy[1, 0], fy[1, 1]]]
    q1 = [[2 * fx[0, 0] / gamma[0], 2 * fx[0, 1] / gamma[0]],
          [2 * fy[0, 0] / gamma[1], 2 * fy[0, 1] / gamma[1]]]

    return numpy.array(q0), numpy.array(q1), x, q_dx, y, q_dy


def chromaticity(ring: Lattice, method: Optional[str] = 'linopt',
                 dpm: Optional[float] = 0.02, npoints: Optional[int] = 11,
                 order: Optional[int] = 3,
                 dp: Optional[float] = 0,
                 **kwargs):
    r"""Computes the non-linear chromaticities

    This function computes the tunes for the specified momentum offsets.
    Then it fits this data and returns the chromaticity up to the given
    order (npoints>order)

    Parameters:
        ring:       Lattice description
        method:     ``'linopt'`` (dfault) returns the tunes from the
          :py:func:`linopt6` function,

          ``'fft'`` tracks a single particle and computes the tunes with fft,

          ``'laskar'`` tracks a single particle and computes the tunes with
          NAFF.
        dpm:        Maximum momentum deviation
        npoints:    Number of momentum deviations
        order:      Order of the fit
        dp:         Central momentum deviation

    Returns:
        fit (ndarray): Horizontal Vertical polynomial coefficients
                       from ``numpy.polyfit`` of shape (2, order+1)
        dpa: dp array of shape (npoints,)
        qz: tune array of shape (npoints, 2)
    """
    if order == 0:
        return get_chrom(ring, dp=dp, **kwargs)
    elif order > npoints - 1:
        raise ValueError('order should be smaller than npoints-1')
    else:
        dpa = numpy.linspace(-dpm, dpm, npoints)
        qz = []
        for dpi in dpa:
            qz.append(get_tune(ring, method=method, dp=dp+dpi, remove_dc=True,
                      **kwargs))
        fit = numpy.polyfit(dpa, qz, order)[::-1]
        fitx = fit[:, 0]/factorial(numpy.arange(order + 1))
        fity = fit[:, 1]/factorial(numpy.arange(order + 1))
        return numpy.array([fitx, fity]), dpa, numpy.array(qz)


def gen_detuning_elem(ring: Lattice, orbit: Optional[Orbit] = None) -> Element:
    """Generates a detuning element

    Parameters:
        ring:           Lattice description
        orbit:          Avoids looking for initial the closed orbit if it is
          already known ((6,) array).

    Returns:
        detuneElem:     Element reproducing the detuning of the ring with
          amplitude and momentum
    """
    if orbit is None:
        orbit, _ = find_orbit(ring)
    lindata0, _, _ = ring.linopt6(get_chrom=False, orbit=orbit)
    xsi = get_chrom(ring.radiation_off(copy=True))
    r0, r1, x, q_dx, y, q_dy = detuning(ring.radiation_off(copy=True),
                                        xm=1.0e-4, ym=1.0e-4, npoints=3)
    nonlin_elem = Element('NonLinear', PassMethod='DeltaQPass',
                          Betax=lindata0.beta[0], Betay=lindata0.beta[1],
                          Alphax=lindata0.alpha[0], Alphay=lindata0.alpha[1],
                          Qpx=xsi[0], Qpy=xsi[1],
                          A1=r1[0][0], A2=r1[0][1], A3=r1[1][1],
                          T1=-orbit, T2=orbit)
    return nonlin_elem



"""Access to properties of RF cavities"""
from enum import Enum
import numpy
from typing import Optional
import functools
from .elements import RFCavity
from .utils import AtError, make_copy, Refpts, get_bool_index
from .lattice_object import Lattice

__all__ = ['frequency_control',
           'get_rf_frequency', 'get_rf_voltage', 'set_rf_voltage',
           'get_rf_timelag', 'set_rf_timelag', 'set_cavity', 'Frf']


def frequency_control(func):
    r""" Function to be used as decorator for
    :pycode:`func(ring, *args, **kwargs)`

    If :pycode:`ring.is_6d` is :py:obj:`True` **and** *dp*, *dct* or *df*
    is specified in *kwargs*, make a copy of *ring* with a modified
    RF frequency, remove *dp*, *dct* or *df* from *kwargs* and call
    *func* with the modified *ring*.

    If :pycode:`ring.is_6d` is :py:obj:`False` **or** no *dp*, *dct* or
    *df* is specified in *kwargs*, *func* is called unchanged.

    Examples:

        .. code-block:: python

            @frequency_control
            def func(ring, *args, dp=None, dct=None, **kwargs):
                pass
    """

    @functools.wraps(func)
    def wrapper(ring, *args, **kwargs):
        if ring.is_6d:
            momargs = {}
            for key in ['dp', 'dct', 'df']:
                v = kwargs.pop(key, None)
                if v is not None:
                    momargs[key] = v
            if len(momargs) > 0:
                frequency = ring.get_revolution_frequency(**momargs) \
                            * ring.harmonic_number
                ring = set_cavity(ring, Frequency=frequency, copy=True)
        return func(ring, *args, **kwargs)

    return wrapper


class Frf(Enum):
    """Enum class for frequency setting"""
    NOMINAL = 'nominal'


def _select_cav(ring: Lattice, cavpts):
    """Select the cavities"""
    if cavpts is None:
        try:
            cavpts = ring.cavpts
        except AttributeError:
            cavpts = get_bool_index(ring, RFCavity)
    return cavpts


def _singlev(values, attr):
    """Return the single value"""
    vals = numpy.unique(values)
    if len(vals) > 1:
        raise AtError('{0} not equal for all cavities'.format(attr))
    return vals[0]


# noinspection PyUnusedLocal
def _sumv(values, attr):
    """Return the sum of values"""
    return numpy.sum(values)


def _fundmask(ring: Lattice, cavpts):
    freqs = numpy.array([cav.Frequency for cav in ring.select(cavpts)])
    if len(freqs) < 1:
        raise AtError('No cavity found in the lattice')
    mask = (freqs == min(freqs))
    return mask


def _get_cavity(ring: Lattice, attr, fsingle, cavpts=None, array=False):
    cavpts = _select_cav(ring, cavpts)
    vcell = numpy.array([getattr(cav, attr) for cav in ring.select(cavpts)])
    if array:
        return vcell
    else:
        fundmask = _fundmask(ring, cavpts)
        if fundmask is not None:
            vcell = vcell[fundmask]
        return fsingle(vcell, attr)


def get_rf_frequency(ring: Lattice, **kwargs) -> float:
    """Return the RF frequency

    Parameters:
        ring:           Lattice description

    Keyword Arguments:
        cavpts=None:    Cavity location. If None, look for ring.cavpts,
                          otherwise take all cavities.
        array=False:    If False, return the frequency of the selected cavities
                          with the lowest frequency.

                          If True, return the frequency of all selected cavities

    Returns:
        rf_freq:    RF frequency
    """
    return _get_cavity(ring, 'Frequency', _singlev, **kwargs)


def get_rf_voltage(ring: Lattice, **kwargs) -> float:
    """Return the total RF voltage (full ring)

    Parameters:
        ring:           Lattice description

    Keyword Arguments:
        cavpts=None:    Cavity location. If None, look for ring.cavpts,
                          otherwise take all cavities.
        array=False:    If False, return the frequency of the selected cavities
                          with the lowest frequency.

                          If True, return the frequency of all selected cavities

    Returns:
        rf_v:       Total RF voltage (full ring)
    """
    vcell = _get_cavity(ring, 'Voltage', _sumv, **kwargs)
    return ring.periodicity * vcell


def get_rf_timelag(ring: Lattice, **kwargs) -> float:
    """Return the RF time lag

    Parameters:
        ring:           Lattice description

    Keyword Arguments:
        cavpts=None:    Cavity location. If None, look for ring.cavpts,
                          otherwise take all cavities.
        array=False:    If False, return the frequency of the selected cavities
                          with the lowest frequency.

                          If True, return the frequency of all selected cavities

    Returns:
        rf_timelag: RF time lag
    """
    return _get_cavity(ring, 'TimeLag', _singlev, **kwargs)


def set_rf_voltage(ring: Lattice, voltage: float, **kwargs):
    """Set the RF voltage for the full ring

    Parameters:
        ring:           Lattice description
        voltage:        Total RF voltage (full ring)

    Keyword Arguments:
        cavpts=None:    Cavity location. If None, look for ring.cavpts,
                          otherwise take all cavities.
        array=False:    If False, return the frequency of the selected cavities
                          with the lowest frequency.

                          If True, return the frequency of all selected cavities
        copy=False:     If True, returns a shallow copy of ``ring`` with new
                          cavity elements. Otherwise, modify ``ring`` in-place.
    """
    return set_cavity(ring, Voltage=voltage, **kwargs)


def set_rf_timelag(ring: Lattice, timelag: float, **kwargs):
    """Set the RF time lag

    Parameters:
        ring:           Lattice description
        timelag:        RF time lag

    Keyword Arguments:
        cavpts=None:    Cavity location. If None, look for ring.cavpts,
                          otherwise take all cavities.
        array=False:    If False, return the frequency of the selected cavities
                          with the lowest frequency.

                          If True, return the frequency of all selected cavities
        copy=False:     If True, returns a shallow copy of ``ring`` with new
                          cavity elements. Otherwise, modify ``ring`` in-place.
    """
    return set_cavity(ring, TimeLag=timelag, **kwargs)


# noinspection PyPep8Naming
def set_cavity(ring: Lattice, Voltage: Optional[float] = None,
               Frequency: Optional[float] = None,
               TimeLag: Optional[float] = None,
               cavpts: Optional[Refpts] = None,
               copy: Optional[bool] = False,
               array: Optional[bool] = False):
    """
    Set the parameters of the RF cavities

    Parameters:
        ring:       lattice description
        Frequency:  RF frequency [Hz]
        Voltage:    RF voltage [V]
        TimeLag:    RF time shift [-ct]
        cavpts:     Cavity location. If None, look for ring.cavpts, or
                      otherwise take all cavities
        array:      If False, the value is applied as described for
                      set_rf_voltage, set_rf_timelag and set_rf_frequency
                      If True, directly apply the value to the selected
                      cavities. The value must be broadcastable to the number
                      of cavities.
        copy:       If True, returns a shallow copy of ring with new
                      cavity elements. Otherwise, modify ring in-place
    """
    # noinspection PyShadowingNames
    def getv(ring: Lattice, attr, refpts):
        values = numpy.array([getattr(c, attr) for c in ring.select(refpts)])
        valfund = values[fundmask]
        return values, valfund

    cavpts = _select_cav(ring, cavpts)
    fundmask = None if array else _fundmask(ring, cavpts)

    modif = {}
    if Frequency is not None:
        if Frequency is Frf.NOMINAL:
            Frequency = ring.revolution_frequency * ring.harmonic_number
        if fundmask is not None:
            vall, vmain = getv(ring, 'Frequency', cavpts)
            vmain = vmain[0]
            if vmain == 0.0:
                vall[fundmask] = Frequency
                Frequency = vall
            else:
                Frequency *= vall/vmain
        modif['Frequency'] = Frequency

    if TimeLag is not None:
        if fundmask is not None:
            vall, vmain = getv(ring, 'TimeLag', cavpts)
            TimeLag += vall-_singlev(vmain, 'TimeLag')
        modif['TimeLag'] = TimeLag

    if Voltage is not None:
        vcell = Voltage / ring.periodicity
        if fundmask is not None:
            vall, vmain = getv(ring, 'Voltage', cavpts)
            vmain = numpy.sum(vmain)
            if vmain == 0.0:
                vall[fundmask] = vcell/numpy.count_nonzero(fundmask)
                vcell = vall
            else:
                vcell *= vall/vmain
        modif['Voltage'] = vcell

    # noinspection PyShadowingNames
    @make_copy(copy)
    def apply(ring: Lattice, cavpts, modif):
        ncavs = ring.refcount(cavpts)
        for attr in modif.keys():
            try:
                values = numpy.broadcast_to(modif[attr], (ncavs,))
            except ValueError:
                raise AtError('set_cavity args should be either scalar or '
                              'a ({0},) vector'.format(ncavs))
            for val, cavity in zip(values, ring.select(cavpts)):
                cavity.update({attr: val})

    return apply(ring, cavpts, modif)


Lattice.get_rf_voltage = get_rf_voltage
Lattice.get_rf_frequency = get_rf_frequency
Lattice.get_rf_timelag = get_rf_timelag
Lattice.set_rf_voltage = set_rf_voltage
Lattice.set_rf_timelag = set_rf_timelag
Lattice.set_cavity = set_cavity
Lattice.rf_voltage = property(get_rf_voltage, set_rf_voltage,
                              doc="RF voltage of the full ring [V]")
Lattice.rf_timelag = property(get_rf_timelag, set_rf_timelag,
                              doc="Time lag of the fundamental mode [m]")



import sys
import numpy as np
import at
from at.collective import Wake, ResWallElement, WakeComponent
from mpi4py import MPI
import pickle as pkl


class Logger(object):
    def __init__(self, fName):

        self.terminal = sys.stdout
        self.log = open(fName, "a")

    def write(self, message):
        self.terminal.write(message)
        self.log.write(message)

    def flush(self):
        # this flush method is needed for python 3 compatibility.
        # this handles the flush command by doing nothing.
        # you might want to specify some extra behavior here.
        pass


def launch():

    sys.stdout = Logger('./logfile.txt')
    sys.stderr = Logger('./logfile.txt')

    comm = MPI.COMM_WORLD
    size = comm.Get_size()
    rank = comm.Get_rank()

    print(size, rank)

    # Set up the ring
    ring = at.load_m('../../../machine_data/esrf.m')
    ring.disable_6d(cavity_pass='RFCavityPass')

    current = 200e-3
    nturns = 10000
    Nbunches = 992
    Npart = int(1 * Nbunches)

    bunchSpacing = ring.circumference/Nbunches
    ring.beam_current = current

    ring.set_fillpattern(Nbunches)

    ring.set_rf_frequency()
    freqres = ring.rf_frequency

    allSextInds = at.get_refpts(ring, at.elements.Sextupole)
    sdInds = []
    sfInds = []
    for i in allSextInds:
        if ring[i].PolynomB[2] > 0:
            sfInds.append(i)
        else:
            sdInds.append(i)
    sdInds = np.array(sdInds)
    sfInds = np.array(sfInds)
      
    at.fit_chrom(ring, sfInds, sdInds, [0, 0])

    l0, _, _ = ring.get_optics()

    fring, _ = at.fast_ring(ring)

    # Define the resonator parameters and current
    wturns = 30
    srange = Wake.build_srange(1e-4, 0.3, 1.0e-4, 1.0e-1, bunchSpacing,
                               ring.circumference * wturns)

    sigmac = 1e8
    rvac = 15e-3
    beta = 1
    welem = ResWallElement('wake', ring, srange, WakeComponent.DX,
                           ring.circumference, rvac, sigmac, beta,
                           Nslice=1, Nturns=wturns)
    fring.append(welem)

    bmon = at.BeamMoments('mon')
    fring.append(bmon)

    # Particle generation and tracking
    sigm = at.sigma_matrix(ring.enable_6d(copy=True))
    part = at.beam(Npart, sigm)
    part[0, :] = 1e-6
    part[1, :] = 0

    fring.track(part, nturns=nturns, refpts=None, in_place=True)

    x_all = bmon.means[0, :, :]
    xp_all = bmon.means[1, :, :]
    if rank == 0:
        outdict = {'x': x_all.T, 'xp': xp_all.T,
                   'ring': ring, 'sigmac': sigmac, 'rvac': rvac,
                   'M': Nbunches, 'alphax': l0['alpha'][0],
                   'betax': l0['beta'][0], 'current': current}
        pkl.dump(outdict, open('./TRW_output.pkl', 'wb'))


if __name__ == '__main__':
    launch()



#!/usr/bin/env python3

import at
import numpy
from mpi_sweep import mpi_sweep

# this code will run on all MPI nodes
d1 = at.elements.Drift('DR01', 3)
qf = at.elements.Quadrupole('QF', 1, 0.2)
d2 = at.elements.Drift('DR02', 3)
qd = at.elements.Quadrupole('QD', 1, -0.2)
fodocell = numpy.array([d1, qf, d2, qd])
thering = at.lattice.Lattice(fodocell, energy=1e6)

def generator():
    """
    Generate input parameter list for worker function.
    It will only be executed on node 0 to save on allocations.
    """
    return [(0.1,-0.1), (0.2,-0.2), (0.3,-0.3)]

def worker(arg, **kwargs):
    """
    Function to run on multiple MPI nodes with argument arg provided
    by the generator function.
    This function may run multiple times on one node.
    """
    # additional arguments may be used
    # kwargs['name']  # name of the processor
    # kwargs['rank']  # rank (id) of the node
    # kwargs['size']  # total number of nodes
    qf.K = arg[0]
    qd.K = arg[1]
    return at.find_m44(thering)[0]

def collector(args):
    """
    Collect computed data.
    This function will only be executed on node 0.
    """
    with open('output.csv', 'w') as f:
        f.write('Kqf,Kqd')
        [f.write(f',m{i}{j}') for i in range(0, 4) for j in range(0, 4)]
        f.write('\n')
        for case in args:
            params, output = case
            f.write(f'{params[0]},{params[1]}')
            [f.write(f',{output[i,j]}') for i in range(0, 4) for j in range(0, 4)]
            f.write('\n')

mpi_sweep(generator, worker, collector)



"""
Handling of JSON files
"""

from __future__ import annotations

__all__ = ["save_json", "load_json"]

from os.path import abspath
import json
from typing import Optional, Any

import numpy as np

from .allfiles import register_format
from .utils import keep_elements, keep_attributes
from ..lattice import Element, Lattice, Particle


class _AtEncoder(json.JSONEncoder):
    """JSON encoder for specific AT types"""

    def default(self, obj):
        if isinstance(obj, Element):
            return obj.to_dict()
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        elif isinstance(obj, Particle):
            return obj.to_dict()
        else:
            return super().default(obj)


def save_json(
    ring: Lattice, filename: Optional[str] = None, compact: bool = False
) -> None:
    """Save a :py:class:`.Lattice` as a JSON file

    Parameters:
        ring:           Lattice description
        filename:       Name of the JSON file. Default: outputs on
          :py:obj:`sys.stdout`
        compact:        If :py:obj:`False` (default), the JSON file is pretty-printed
          with line feeds and indentation. Otherwise, the output is a single line.

    See Also:
        :py:func:`.save_lattice` for a generic lattice-saving function.
        :py:meth:`.Lattice.save` for a generic lattice-saving method.
    """
    indent = None if compact else 2
    data = dict(
        atjson=1, elements=list(keep_elements(ring)), properties=keep_attributes(ring)
    )
    if filename is None:
        print(json.dumps(data, cls=_AtEncoder, indent=indent))
    else:
        with open(filename, "wt") as jsonfile:
            json.dump(data, jsonfile, cls=_AtEncoder, indent=indent)


def load_json(filename: str, **kwargs) -> Lattice:
    """Create a :py:class:`.Lattice`  from a JSON file

    Parameters:
        filename:           Name of a JSON file

    Keyword Args:
        *:                  All keywords update the lattice properties

    Returns:
        lattice (Lattice):          New :py:class:`.Lattice` object

    See Also:
        :py:meth:`.Lattice.load` for a generic lattice-loading method.
    """

    def json_generator(params: dict[str, Any], fn):

        with open(params.setdefault("in_file", fn), "rt") as jsonfile:
            data = json.load(jsonfile)
        # Check the file signature - For later use
        try:
            atjson = data["atjson"]  # noqa F841
        except KeyError:
            atjson = 1  # noqa F841
        # Get elements
        elements = data["elements"]
        # Get lattice properties
        try:
            properties = data["properties"]
        except KeyError:
            properties = {}
        particle_dict = properties.pop("particle", {})
        params.setdefault("particle", Particle(**particle_dict))

        for k, v in properties.items():
            params.setdefault(k, v)
        for idx, elem_dict in enumerate(elements):
            yield Element.from_dict(elem_dict, index=idx, check=False)

    return Lattice(abspath(filename), iterator=json_generator, **kwargs)


register_format(
    ".json",
    load_json,
    save_json,
    descr="JSON representation of a python AT Lattice. See :py:func:`.load_json`.",
)



"""A-matrix construction"""
import numpy
from scipy.linalg import block_diag, eig, inv, solve
from math import pi
from at.lattice import AtError

__all__ = ['a_matrix', 'amat', 'jmat', 'jmatswap',
           'get_tunes_damp', 'get_mode_matrices', 'symplectify']

_i2 = numpy.array([[-1.j, -1.], [1., 1.j]])

# Prepare symplectic identity matrix
_j2 = numpy.array([[0., 1.], [-1., 0.]])
_jm = [_j2, block_diag(_j2, _j2), block_diag(_j2, _j2, _j2)]
_jmswap = [_j2, block_diag(_j2, _j2), block_diag(_j2, _j2, _j2.T)]

_vxyz = [_i2, block_diag(_i2, _i2), block_diag(_i2, _i2, _i2)]
_submat = [slice(0, 2), slice(2, 4), slice(4, 6)]


def jmat(ind: int):
    """antisymetric block diagonal matrix [[0, 1][-1, 0]]

    Parameters:
        ind:    Matrix dimension: 1, 2 or 3

    Returns:
        S:      Block diagonal matrix, (2, 2) or (4, 4) or (6, 6)
    """
    return _jm[ind - 1]


def jmatswap(ind: int):
    """Modified antisymetric block diagonal matrix to deal with the swap of the
    longitudinal coordinates
    """
    return _jmswap[ind - 1]


# noinspection PyPep8Naming
def a_matrix(M):
    r"""Find the :math:`\mathbf{A}` matrix from one turn map :math:`\mathbf{M}`

    :math:`\mathbf{A}` represents a change of referential which converts the
    one-turn transfer matrix :math:`\mathbf{M}` into a set of rotations:

    .. math:: \mathbf{M} = \mathbf{A} \cdot \mathbf{R} \cdot \mathbf{A}^{-1}

    with :math:`\mathbf{R}` A block-diagonal matrix:

    .. math::

        \mathbf{R}=\begin{pmatrix}rot_1 & 0 & 0 \\ 0 & rot_2 & 0 \\ 0 & 0 &
        rot_3\end{pmatrix} \text{, and } rot_i = \begin{pmatrix}\cos{\mu_i} &
        \sin{\mu_i} \\ -\sin{\mu_i} & cos{\mu_i}\end{pmatrix}

    With radiation damping, the diagonal blocks are instead damped rotations:

    .. math::

        rot_i = \begin{pmatrix}\exp{(-\alpha_i}) & 0 \\ 0 & \exp{(-\alpha_i)}
        \end{pmatrix} \cdot \begin{pmatrix}\cos{\mu_i} & \sin{\mu_i} \\
        -\sin{\mu_i} & cos{\mu_i}\end{pmatrix}

    The order of diagonal blocks it set so that it is close to the order of
    x,y,z.

    Parameters:
        M:     (m, m)  transfer matrix for 1 turn

    m, the dimension of :math:`\mathbf{M}`, may be 2 (single plane),
    4 (betatron motion) or 6 (full motion)


    Returns:
        A:      (m, m)  A-matrix
        eigval: (m/2,)    Vector of Eigen values of T

    References:
        **[1]** Etienne Forest, Phys. Rev. E 58, 2481 – Published 1 August 1998
    """
    nv = M.shape[0]
    dms = int(nv / 2)
    jmt = jmatswap(dms)
    select = numpy.arange(0, nv, 2)
    rbase = numpy.stack((select, select), axis=1).flatten()

    # noinspection PyTupleAssignmentBalance
    lmbd, vv = eig(M)
    # Compute the norms
    vp = vv.conj().T @ jmt
    n = -0.5j * numpy.sum(vp.T * vv, axis=0)
    if any(abs(n) < 1.0E-12):
        raise AtError('Unstable ring')
    # Move positive before negatives
    order = rbase + (n < 0)
    vv = vv[:, order]
    n = n[order]
    lmbd = lmbd[order]
    # Normalize vectors
    vn = vv / numpy.sqrt(abs(n)).reshape((1, nv))
    # find the vectors that project most onto x,y,z, and reorder
    # nn will have structure
    #  n1x n1y n1z
    #  n2x n2y n2z
    #  n3x n3y n3z
    nn = 0.5 * abs(numpy.sqrt(-1.j * vn.conj().T @ jmt @ _vxyz[dms - 1]))
    rows = list(select)
    order = []
    for ixz in select:
        ind = numpy.argmax(nn[rows, ixz])
        order.append(rows[ind])
        del rows[ind]
    v_ordered = vn[:, order]
    lmbd = lmbd[order]
    aa = numpy.vstack((numpy.real(v_ordered), numpy.imag(v_ordered))).reshape(
        (nv, nv), order='F')
    return aa, lmbd


# noinspection PyPep8Naming
def amat(M):
    """Find the A matrix from one turn map matrix T

    Provided for backward compatibility, see :py:func:`a_matrix`

    Parameters:
        M:     (m, m)  transfer matrix for 1 turn

    Returns:
        A:     (m, m)  A-matrix
    """
    aa, _ = a_matrix(M)
    return aa


# noinspection PyPep8Naming
def symplectify(M):
    """Makes A matrix more symplectic

    following the Healy algorithm described by MacKay

    Parameters:
        M:  Almost symplectic matrix

    Returns:
        MS: Symplectic matrix

    References:
        **[1]** `W.W.MacKay, Comment on Healy's symplectification algorithm,
        Proceedings of EPAC 2006
        <https://accelconf.web.cern.ch/e06/PAPERS/WEPCH152.PDF>`_
    """
    nv = M.shape[0]
    S = jmat(nv // 2)
    I = numpy.identity(nv)

    V = S @ (I - M) @ inv(I + M)
    # V should be almost symmetric.  Replace with symmetrized version.
    W = (V + V.T) / 2
    # Now reconstruct M from W
    SW = S @ W
    MS = (I + SW) @ inv(I - SW)
    return MS


# noinspection PyPep8Naming
def get_mode_matrices(A):
    """Derives the R-matrices from the A-matrix

    Parameters:
        A:  A-matrix

    Returns:
        R:

    References:
        **[1]** Andrzej Wolski, Phys. Rev. ST Accel. Beams 9, 024001 –
        Published 3 February 2006
    """

    def mul2(slc):
        return A[:, slc] @ tt[slc, slc]

    dms = A.shape[0] // 2
    # Rk = A * Ik * A.T                     Only for symplectic
    # modelist = [numpy.dot(A[:, sl], A.T[sl, :]) for sl in _submat[:dms]]
    # Rk = A * S * Ik * inv(A) * S.T        Even for non-symplectic
    ss = jmat(dms)
    tt = jmatswap(dms)
    a_s = numpy.concatenate([mul2(slc) for slc in _submat[:dms]], axis=1)
    inva = solve(A, ss.T)
    modelist = [a_s[:, sl] @ inva[sl, :] for sl in _submat[:dms]]
    return numpy.stack(modelist, axis=0)


# noinspection PyPep8Naming
def get_tunes_damp(M, R=None):
    r"""Computes the mode emittances, tunes and damping times

    Parameters:
        M:     (m, m) transfer matrix for 1 turn
        R:     (m, m) beam matrix (optional), allows computing the mode
                emittances

    m, the dimension of :math:`\mathbf{M}`, may be 2 (single plane),
    4 (betatron motion) or 6 (full motion)

    Returns:
        V:     record array with the following fields:

          **tunes**             (m/2,) tunes of the m/2 normal modes

          **damping_rates**     (m/2,) damping rates of the m/2 normal modes

          **mode_matrices**     (m/2, m, m) the R-matrices of the m/2 normal
          modes

          **mode_emittances**   Only if R is specified: (m/2,) emittance of each
          of the m/2 normal modes
    """
    nv = M.shape[0]
    dms = int(nv / 2)
    A, vps = a_matrix(M)
    tunes = numpy.mod(numpy.angle(vps) / 2.0 / pi, 1.0)
    damping_rates = -numpy.log(numpy.absolute(vps))

    if R is None:
        return numpy.rec.fromarrays(
            (numpy.array(tunes), numpy.array(damping_rates),
             numpy.array(get_mode_matrices(A))),
            dtype=[('tunes', numpy.float64, (dms,)),
                   ('damping_rates', numpy.float64, (dms,)),
                   ('mode_matrices', numpy.float64, (dms, nv, nv))]
        )
    else:
        jmt = jmat(dms)
        rdiag = numpy.diag(A.T @ jmt @ R @ jmt @ A)
        mode_emit = -0.5 * (rdiag[0:nv:2] + rdiag[1:nv:2])
        return numpy.rec.fromarrays(
            (numpy.array(tunes), numpy.array(damping_rates),
             numpy.array(get_mode_matrices(A)), mode_emit),
            dtype=[('tunes', numpy.float64, (dms,)),
                   ('damping_rates', numpy.float64, (dms,)),
                   ('mode_matrices', numpy.float64, (dms, nv, nv)),
                   ('mode_emittances', numpy.float64, (dms,))]
        )



import pytest
import numpy
from at.tracking.atpass import atpass
from at import elements, uint32_refpts


def test_incorrect_types_raises_value_error(rin):
    with pytest.raises(TypeError):
        atpass(1, rin, 1)
    with pytest.raises(TypeError):
        atpass([], 1, 1)
    with pytest.raises(TypeError):
        atpass([], rin, 'a')


def test_incorrect_dimensions_raises_value_error():
    rin = numpy.array(numpy.zeros((7, 1)))
    with pytest.raises(ValueError):
        atpass([], rin, 1)
    rin = numpy.array(numpy.zeros((1, 6)))
    with pytest.raises(ValueError):
        atpass([], rin, 1)


def test_c_aligned_array_raises_value_error():
    rin = numpy.zeros((2, 6))
    with pytest.raises(ValueError):
        atpass([], rin, 1)


def test_transposed_array_raises_value_error():
    rin_transposed = numpy.asfortranarray(numpy.zeros((2, 6)))
    with pytest.raises(ValueError):
        atpass([], rin_transposed.T, 1)


def test_transposed_c_array_does_not_raise_value_error():
    rin_c = numpy.zeros((2, 6))
    atpass([], rin_c.T, 1)


def test_transposed_c_array_gives_same_result_as_fortran_array():
    """
    This test explores the behaviour of C- and Fortran- aligned arrays.
    Since we use the data from the numpy array directly, some of the
    behaviour we see is not what we'd expect if we used numpy as intended.
    """
    # Standard numpy array (6, 2) as used in pyAT.
    rin_fortran = numpy.array(numpy.arange(12).reshape(6, 2), order='F') * 1e-5
    # Taking an copy of the transpose and making sure it is C-aligned should
    # give us the same layout of data in memory, but with a (2, 6) array.
    rin_c = numpy.copy(rin_fortran.T, order='C')
    lat = [elements.Drift('drift', 1.0)]
    rout_fortran = atpass(lat, rin_fortran, 1)
    # at.c does not accept (x, 6) arrays.  This transpose allows rin_c
    # to pass the dimension check, but does NOT change the layout in memory
    # since in Python it returns a 'view' on the array.
    # The AT C code would give the same result for rin_c without the
    # transpose if the dimension check were not there.
    rin_c_transposed = rin_c.T
    rout_c = atpass(lat, rin_c_transposed, 1)
    numpy.testing.assert_equal(rout_c, rout_fortran)


def test_missing_pass_method_raises_attribute_error(rin):
    lat = [elements.Marker('marker')]
    del lat[0].PassMethod
    with pytest.raises(AttributeError):
        atpass(lat, rin, 1)


def test_missing_length_raises_attribute_error(rin):
    lat = [elements.Drift('drift', 1.0)]
    del lat[0].Length
    with pytest.raises(AttributeError):
        atpass(lat, rin, 1)


@pytest.mark.parametrize("reuse", (True, False))
def test_reuse_attributes(rin, reuse):
    lat = [elements.Drift('drift', 1.0)]
    rin[0, 0] = 1e-6
    rin[1, 0] = 1e-6
    rin_copy = numpy.copy(rin)
    # two turns with original lattice
    atpass(lat, rin, 2)
    # one turn with original lattice
    atpass(lat, rin_copy, 1)
    # change an attribute
    lat[0].Length = 2
    # one turn with altered lattice
    atpass(lat, rin_copy, 1, reuse=reuse)
    if reuse:
        numpy.testing.assert_equal(rin, rin_copy)
    else:
        with pytest.raises(AssertionError):
            numpy.testing.assert_equal(rin, rin_copy)


def test_two_particles_for_two_turns():
    rin = numpy.asfortranarray(numpy.zeros((6, 2)))
    lat = [elements.Drift('drift', 1.0)]
    rin[1][0] = 1e-6
    rin[3][0] = -2e-6
    rout = atpass(lat, rin, 2, refpts=uint32_refpts([1], 1))
    # results from Matlab
    rout_particle1_turn1 = numpy.array([1e-6, 1e-6, -2e-6, -2e-6, 0,
                                        2.5e-12]).reshape(6, 1)
    rout_particle1_turn2 = numpy.array([2e-6, 1e-6, -4e-6, -2e-6, 0,
                                        5e-12]).reshape(6, 1)
    # the second particle doesn't change
    rout_particle2 = numpy.zeros((6, 1))
    # the second index is particle number
    numpy.testing.assert_equal(rout[:, 0, :, 0], rout_particle1_turn1)
    numpy.testing.assert_equal(rout[:, 1, :, 0], rout_particle2)
    # the fourth index is turn number
    numpy.testing.assert_equal(rout[:, 0, :, 1], rout_particle1_turn2)
    numpy.testing.assert_equal(rout[:, 1, :, 1], rout_particle2)


def test_one_particle_for_two_turns_with_no_refpts(rin):
    lat = [elements.Drift('drift', 1.0)]
    rin[1][0] = 1e-6
    rin[3][0] = -2e-6
    atpass(lat, rin, 2)
    rout_expected = numpy.array([2e-6, 1e-6, -4e-6, -2e-6, 0,
                                 5e-12]).reshape(6, 1)
    # rin is changed in place
    numpy.testing.assert_equal(rin, rout_expected)


def test_1d_particle():
    # This is just a demonstration that if you pass no refpts you get back
    # a (6, *, 0, *) array. You may do this if you want only to operate
    # on rin in-place.
    lat = [elements.Drift('drift', 1.0)]
    rin = numpy.zeros(6,)
    rin[1] = 1e-6
    # an empty refpts returns only the value at the end of the last turn
    rout = atpass(lat, rin, 1)
    # output shape: (dimensions, nparticles, refpts, nturns)
    # if no refpts are supplied, the output is (6, 1, 0, 1) and contains
    # no data
    assert rout.shape == (6, 1, 0, 1)
    rout_expected = numpy.array([1e-6, 1e-6, 0, 0, 0, 5e-13])
    # rin is changed in place
    numpy.testing.assert_equal(rin, rout_expected)



"""Description of particles"""
from ..constants import e_mass, p_mass
import numpy
from warnings import warn
from typing import Optional, Dict


class Particle(object):
    """
    Particle object: it defines the properties of the particles circulating
    in a ring
    """
    _known = dict(
        relativistic=dict(rest_energy=0.0, charge=-1.0),
        electron=dict(rest_energy=e_mass, charge=-1.0),
        positron=dict(rest_energy=e_mass, charge=1.0),
        proton=dict(rest_energy=p_mass, charge=1.0)
    )

    def __init__(self, name: Optional[str] = 'relativistic', **kwargs):
        """

        Parameters:
            name:   Particle name. 'electron', 'positron and 'proton' are
              predefined. For other particles, the rest energy and charge
              must be provided as keywords.

        Keyword Arguments:
            rest_energy:    Particle rest energy [ev]
            charge:         Particle charge [elementary charge]
            *:              Other keywords will be set as attributes of the
                              particle
        """
        if name != 'relativistic':
            warn(UserWarning("AT tracking still assumes beta==1\n"
                             "Make sure your particle is ultra-relativistic"))
        if name in self._known:
            kwargs.update(self._known[name])
        self.name = name
        # Use a numpy scalar to allow division by zero
        self._rest_energy = numpy.array(kwargs.pop('rest_energy'), dtype=float)
        self._charge = kwargs.pop('charge')
        for (key, val) in kwargs.items():
            setattr(self, key, val)

    def to_dict(self) -> Dict:
        attrs = vars(self).copy()
        attrs['rest_energy'] = attrs.pop('_rest_energy')
        attrs['charge'] = attrs.pop('_charge')
        return attrs

    def __repr__(self):
        if self.name in self._known:
            return "Particle('{0}')".format(self.name)
        else:
            attrs = self.to_dict()
            name = attrs.pop('name')
            args = ', '.join('{0}={1!r}'.format(k, v)
                             for k, v in attrs.items())
            return "Particle('{0}', {1})".format(name, args)

    def __str__(self):
        if self.name in self._known:
            return self.name
        else:
            return self.__repr__()

    # Use properties so that they are read-only
    @property
    def rest_energy(self) -> numpy.ndarray:
        """Particle rest energy [eV]"""
        return self._rest_energy

    @property
    def charge(self) -> float:
        """Particle charge [elementary charge]"""
        return self._charge



"""
Functions used to
calculate the loss boundary for different
grid definitions
"""

from at.lattice import Lattice, AtError, AtWarning
from typing import Optional, Sequence
from enum import Enum
import numpy
from scipy.ndimage import binary_dilation, binary_opening
from collections import namedtuple
import time
import warnings

__all__ = ['GridMode']

_pdict = {'x': 0, 'xp': 1,
          'y': 2, 'yp': 3,
          'dp': 4, 'ct': 5}


class GridMode(Enum):
    """
    Grid definition for 2D acceptance boundary search
    """
    RADIAL = 0      #: full [:math:`\:r, \theta\:`] grid
    CARTESIAN = 1   #: full [:math:`\:x, y\:`] grid
    RECURSIVE = 2   #: radial recursive search


def grid_config(planes, amplitudes, npoints, bounds, grid_mode,
                shift_zero):
    """"
    Returns an object that defines the grid configuration
    """
    bounds = numpy.atleast_2d(bounds)
    bounds = tuple(map(tuple, bounds))
    shape = numpy.array(npoints, dtype=numpy.int32)
    d = {'planes': numpy.atleast_1d(planes),
         'planesi': numpy.atleast_1d(get_plane_index(planes)),
         'amplitudes': numpy.atleast_1d(amplitudes),
         'shape': numpy.atleast_1d(shape),
         'bounds': bounds,
         'mode': grid_mode,
         'shift_zero': shift_zero}
    return namedtuple('config', d.keys())(**d)


def grid(grid, offset):
    """"
    Returns a grid object
    """
    d = {'grid': numpy.atleast_2d(grid),
         'offset': numpy.atleast_1d(offset)}
    return namedtuple('grid', d.keys())(**d)


def get_plane_index(planes):
    """
    Converts plane to particle coordinate index
    """
    planesi = numpy.array([], dtype=numpy.int32)
    for i, p in enumerate(numpy.atleast_1d(planes)):
        if isinstance(p, str):
            try:
                planesi = numpy.append(planesi, _pdict[p])
            except KeyError:
                raise AtError('Allowed values for plane are x,xp,y,yp,dp,ct')
        else:
            raise AtError('Allowed values for plane are x,xp,y,yp,dp,ct')
    return planesi


def set_ring_orbit(ring, dp, obspt, orbit):
    """
    Returns a ring starting at obspt and initial
    closed orbit
    """
    if obspt is not None:
        assert numpy.isscalar(obspt), 'Scalar value needed for obspt'
        ring = ring.rotate(obspt)

    if orbit is None:
        orbit = ring.find_orbit(dp=dp)[0]

    return orbit, ring


def grid_configuration(planes, npoints, amplitudes, grid_mode, bounds=None,
                       shift_zero=1.0e-6):
    """
    Return a grid configuration based on user input parameters, the ordering
    is as follows: CARTESIAN: (x,y), RADIAL/RECURSIVE (r, theta).
    Scalar inputs can be used for 1D grid
    """
    ndims = len(numpy.atleast_1d(planes))
    if ndims > 2 or ndims == 0:
        raise AtError('planes can have 1 or 2 element (1D or 2D aperture)')
    elif ndims == 1 and grid_mode is GridMode.RADIAL:
        grid_mode = GridMode.CARTESIAN

    if numpy.shape(numpy.atleast_1d(npoints)) != (ndims,):
        raise AtError('npoints shape should be (len(planes),)')
    if numpy.shape(numpy.atleast_1d(amplitudes)) != (ndims,):
        raise AtError('amplitudes shape should be (len(planes),)')
    if (numpy.shape(numpy.atleast_2d(bounds)) != (ndims, 2)
       and bounds is not None):
        raise AtError('bounds shape should be (len(planes),2)')

    if grid_mode is GridMode.RADIAL or grid_mode is GridMode.RECURSIVE:
        if bounds is None:
            bounds = numpy.array([[0, 1], [numpy.pi, 0]])
        bounds[0][bounds[0] == 0] = 1.0e-6
    elif grid_mode is GridMode.CARTESIAN:
        if bounds is None:
            bounds = numpy.array([[p-1, 1] for p in range(ndims)])
    else:
        raise AtError('GridMode {0} undefined.'.format(grid_mode))

    config = grid_config(planes, amplitudes, npoints,
                         bounds, grid_mode, shift_zero)
    return config


def get_parts(config, offset):
    """
    Generate a (6,n) particles array based on the grid configuration
    and initial offset, returns a grid object containing the grid and
    the offset from which the array can be reconstructed
    """
    def get_part_grid_uniform(bnd, np, amp):
        x = [numpy.linspace(*b, n)*a for a, b, n in zip(amp, bnd, np)]
        try:
            g1, g2 = numpy.meshgrid(*x)
        except ValueError:
            g1, g2 = numpy.meshgrid(x, 0.0)
        return numpy.array([g1.flatten(), g2.flatten()])

    def get_part_grid_radial(bnd, np, amp):
        x = [numpy.linspace(*b, n) for b, n in zip(bnd, np)]
        g1, g2 = numpy.meshgrid(*x)
        g1r = amp[0]*numpy.cos(g2)*g1
        g2r = amp[1]*numpy.sin(g2)*g1
        return numpy.array([g1r.flatten(), g2r.flatten()])

    pind = config.planesi
    amp = config.amplitudes
    np = config.shape
    bnd = config.bounds
    gm = config.mode

    if gm is GridMode.CARTESIAN:
        g = get_part_grid_uniform(bnd, np, amp)
    elif gm is GridMode.RADIAL:
        g = get_part_grid_radial(bnd, np, amp)
    parts = numpy.zeros((6, numpy.prod(np)))
    parts[pind, :] = [g[i] for i in range(len(pind))]
    offset = numpy.array(offset) + config.shift_zero
    parts = (parts.T+offset).T
    return parts, grid(g, offset[pind])


def get_survived(parts, ring, nturns, use_mp, **kwargs):
    """
    Track a grid through the ring and extract survived particles
    """
    _, _, td = ring.track(parts, nturns=nturns, losses=True, use_mp=use_mp, **kwargs)
    return numpy.invert(td['loss_map'].islost)


def get_grid_boundary(mask, grid, config):
    """
    Compute the boundary of survided particles array
    """
    def nearest_order(grid):
        #  keep only max r for each angle
        angle = numpy.arctan2(*grid)
        norm = numpy.linalg.norm(grid.T, axis=1)
        val, inv = numpy.unique(angle, return_inverse=True)
        gf = numpy.zeros((2, len(val)))
        for i, v in enumerate(val):
            inds = numpy.where(inv == i)[0]
            ni = norm[inds]
            nim = numpy.where(ni == numpy.amax(ni))[0]
            ind = inds[nim][0]
            gf[:, i] = grid[:, ind]
        #  first sort by angle
        idx = numpy.argsort(numpy.arctan2(*gf))
        gf = gf[:, idx]
        #  now sort by closest neighbour on normalized grid
        x, y = gf[0, :].copy(), gf[1, :].copy()
        dxmin = min(numpy.diff(numpy.unique(x)))
        dymin = min(numpy.diff(numpy.unique(y)))
        iorder = [0]
        for i in range(1, len(x)):
            xnow = x[iorder[-1]]
            ynow = y[iorder[-1]]
            dd = numpy.sqrt(((x-xnow)/dxmin)**2+((y-ynow)/dymin)**2)
            if i <= 3:
                ic = [j for j in numpy.argsort(dd) if j not in iorder]
            else:
                direction = numpy.sign(iorder[-1]-iorder[-2])
                ic = [j for j in numpy.argsort(dd) if j not in iorder
                      and numpy.sign(j-iorder[-1]) == direction]
            if len(ic) > 0:
                iorder.append(ic[0])
        #  finally connect both ends if distance within unit square
        xnow = x[iorder[-1]]
        ynow = y[iorder[-1]]
        xs = x[iorder[0]]
        ys = y[iorder[0]]
        dd = numpy.sqrt(((xs-xnow)/dxmin)**2+((ys-ynow)/dymin)**2)
        if dd < 1.5:
            iorder.append(iorder[0])
        gf = gf[:, iorder]
        return gf

    def search_bnd(ma, sa):
        bnd = numpy.zeros((2, 1))
        for j, m in enumerate(ma):
            bnd = sa[:, j]
            if not m and j > 0:
                bnd = sa[:, j-1]
                break
        return bnd

    def grid_boundary(mask, grid, config, nclean=2):
        cnt = numpy.flip(config.shape)
        if len(cnt) == 1:
            return vector_boundary(mask, grid)
        mask2d = numpy.reshape(mask.copy(), cnt)
        #  remove isolated points
        for i in range(nclean):
            bnd1 = binary_opening(mask2d, numpy.ones((2, 1)))
            bnd2 = binary_opening(mask2d, numpy.ones((1, 2)))
        bnd = numpy.logical_and(bnd1, bnd2)
        k = numpy.zeros((3, 3), dtype=int)
        k[1] = 1
        k[:, 1] = 1
        bnd = numpy.logical_and(binary_dilation(bnd == 0, border_value=1), bnd)
        bnd = grid.grid[:, bnd.reshape(mask.shape)]
        return nearest_order(bnd)

    def vector_boundary(mask, grid):
        g = grid.grid
        xp, xn = g[0] >= 0, g[0] <= 0
        bp = search_bnd(mask[xp], g[:, xp])
        bn = search_bnd(numpy.flip(mask[xn]),
                        numpy.flip(g[:, xn], axis=1))
        return numpy.squeeze([bn[0], bp[0]])

    def radial_boundary(mask, grid):
        angles = numpy.round(numpy.arctan2(*grid.grid), decimals=8)
        angles, invi = numpy.unique(angles, return_inverse=True)
        bnd = numpy.zeros((2, len(angles)))
        for i in range(len(angles)):
            sa = grid.grid[:, invi == i]
            ma = mask[invi == i]
            bnd[:, i] = search_bnd(ma, sa)
        return bnd

    if not numpy.any(mask):
        msg = ("No particle survived, please check your grid "
               "or lattice. Acceptance set to [0.0, 0.0].")
        warnings.warn(AtWarning(msg))
        cnt = numpy.flip(config.shape)
        if len(cnt) == 1:
            return numpy.zeros(2)
        else:
            return numpy.zeros((2, 1))            

    if config.mode is GridMode.RADIAL:
        return radial_boundary(mask, grid)
    elif config.mode is GridMode.CARTESIAN:
        return grid_boundary(mask, grid, config)
    else:
        raise AtError('GridMode {0} undefined.'.format(grid.mode))


def grid_boundary_search(ring, planes, npoints, amplitudes, nturns=1024,
                         obspt=None, dp=None, offset=None, bounds=None,
                         grid_mode=GridMode.RADIAL, use_mp=False,
                         verbose=True, shift_zero=1.0e-9, **kwargs):
    """
    Search for the boundary by tracking a grid
    """
    offset, newring = set_ring_orbit(ring, dp, obspt,
                                     offset)
    config = grid_configuration(planes, npoints, amplitudes,
                                grid_mode, bounds=bounds,
                                shift_zero=shift_zero)

    if verbose:
        print('\nRunning grid boundary search:')
        if obspt is None:
            print('Element {0}, obspt={1}'.format(ring[0].FamName, 0))
        else:
            print('Element {0}, obspt={1}'.format(ring[obspt].FamName,
                                                  obspt))
        print('The grid mode is {0}'.format(config.mode))
        print('The planes are {0}'.format(config.planes))
        print('Number of steps are {0}'.format(config.shape))
        print('The maximum amplitudes are {0}'.format(config.amplitudes))
        print('The maximum boundaries are {0}'.format(config.bounds))
        print('The initial offset is {0} with dp={1}'.format(offset, dp))

    t0 = time.time()
    parts, grid = get_parts(config, offset)
    mask = get_survived(parts, newring, nturns, use_mp, **kwargs)
    survived = grid.grid[:, mask]
    boundary = get_grid_boundary(mask, grid, config)
    if verbose:
        print('Calculation took {0}'.format(time.time()-t0))
    return boundary, survived, grid.grid


def recursive_boundary_search(ring, planes, npoints, amplitudes, nturns=1024,
                              obspt=None, dp=None, offset=None, bounds=None,
                              use_mp=False, divider=2, verbose=True,
                              shift_zero=1.0e-9, **kwargs):
    """
    Recursively search for the boundary in a given plane and direction (angle)
    """
    def search_boundary(planesi, angles, rtol, rsteps, nturns,
                        offset, use_mp, **kwargs):

        ftol = min(rtol/rsteps)
        cs = numpy.squeeze([numpy.cos(angles), numpy.sin(angles)])
        cs = numpy.around(cs, decimals=9)
        fact = numpy.ones(len(angles))
        survived = numpy.full(len(angles), True)
        part = numpy.zeros((6, len(angles)))
        grid = numpy.array([])
        mask = numpy.array([])

        while numpy.any(survived):
            for i, pi in enumerate(planesi):
                part[pi, survived] += cs[i, survived]*rsteps[i]*fact[survived]
            istracked = numpy.array([not numpy.any([numpy.allclose(p, g,
                                                                   rtol=1.0e-9)
                                    for g in grid.T])
                                    for p in part[planesi].T])
            survived = numpy.array([numpy.any([numpy.allclose(p, m,
                                                              rtol=1.0e-9)
                                   for m in mask.T])
                                   for p in part[planesi].T])
            pt = part[:, istracked]
            grid = numpy.hstack([grid,
                                 pt[planesi]]) if grid.size else pt[planesi]
            ptmp = (pt.T + offset).T
            survived[istracked] = get_survived(ptmp, newring, nturns,
                                               use_mp, **kwargs)
            pm = part[:, numpy.logical_and(istracked, survived)]
            mask = numpy.hstack([mask,
                                 pm[planesi]]) if mask.size else pm[planesi]
            for i in range(len(angles)):
                if not survived[i] and fact[i] > ftol:
                    deltas = cs[:, i]*rsteps[:]*min(1, 2*fact[i])
                    if numpy.any(abs(deltas) > abs(part[planesi, i])):
                        part[planesi, i] = numpy.zeros(len(planesi))
                    else:
                        for j, pi in enumerate(planesi):
                            part[pi, i] -= deltas[j]
                    survived[i] = True
                    fact[i] *= 1/divider

        for i, pi in enumerate(planesi):
            part[pi] -= cs[i]*rsteps[i]*fact

        p = numpy.squeeze(part[planesi])
        return p, mask, grid

    offset, newring = set_ring_orbit(ring, dp, obspt, offset)
    config = grid_configuration(planes, npoints, amplitudes,
                                GridMode.RECURSIVE, bounds=bounds,
                                shift_zero=shift_zero)
    rtol = min(numpy.atleast_1d(config.amplitudes/config.shape))
    rstep = config.amplitudes
    if len(numpy.atleast_1d(config.shape)) == 2:
        angles = numpy.linspace(*config.bounds[1], config.shape[1])
    else:
        angles = numpy.linspace(*config.bounds[1], 2)
    angles = numpy.atleast_1d(angles)

    if verbose:
        print('\nRunning recursive boundary search:')
        if obspt is None:
            print('Element {0}, obspt={1}'.format(ring[0].FamName, 0))
        else:
            print('Element {0}, obspt={1}'.format(ring[obspt].FamName,
                                                  obspt))
        print('The grid mode is {0}'.format(config.mode))
        print('The planes are {0}'.format(config.planes))
        print('Number of angles is {0} from {1} to {2} rad'.format(len(angles),
              angles[0], angles[-1]))
        print('The resolution of the search is {0}'.format(rtol))
        print('The initial step size is {0}'.format(rstep))
        print('The initial offset is {0} with dp={1}'.format(offset, dp))

    t0 = time.time()
    result = search_boundary(config.planesi, angles, rtol, rstep,
                             nturns, offset, use_mp, **kwargs)
    if verbose:
        print('Calculation took {0}'.format(time.time()-t0))
    return result


def boundary_search(ring: Lattice, planes, npoints, amplitudes,
                    nturns: Optional[int] = 1024,
                    obspt: Optional[int] = None, dp: Optional[float] = None,
                    offset: Sequence[float] = None, bounds=None,
                    grid_mode: Optional[GridMode] = GridMode.RADIAL,
                    use_mp: Optional[bool] = False,
                    verbose: Optional[bool] = True,
                    shift_zero: Optional[float] = 1.0e-9,
                    **kwargs):
    """
    Computes the loss boundary at a single point in the machine
    """
    divider = kwargs.pop('divider', 2)
    if grid_mode is GridMode.RECURSIVE:
        result = recursive_boundary_search(ring, planes, npoints, amplitudes,
                                           nturns=nturns, obspt=obspt, dp=dp,
                                           offset=offset, bounds=bounds,
                                           use_mp=use_mp, verbose=verbose,
                                           divider=divider,
                                           shift_zero=shift_zero,
                                           **kwargs)
    else:
        result = grid_boundary_search(ring, planes, npoints, amplitudes,
                                      nturns=nturns, obspt=obspt, dp=dp,
                                      offset=offset, bounds=bounds,
                                      grid_mode=grid_mode, use_mp=use_mp,
                                      verbose=verbose, shift_zero=shift_zero,
                                      **kwargs)
    return result



"""
Wake object creation
"""
import numpy
import warnings
from enum import Enum
from scipy.interpolate import interp1d
from ..lattice import AtWarning, AtError
from .wake_functions import long_resonator_wf, transverse_resonator_wf
from .wake_functions import transverse_reswall_wf


class WakeType(Enum):
    """Enum class for wake type"""
    FILE = 1        #: Import from file
    TABLE = 2       #: Provide vectors
    RESONATOR = 3   #: Analytical resonator
    RESWALL = 4     #: Analytical resistive wall


class WakeComponent(Enum):
    """Enum class for wake component"""
    DX = 1          #: Dipole X
    DY = 2          #: Dipole Y
    QX = 3          #: Quadrupole X
    QY = 4          #: Quadrupole Y
    Z = 5           #: Longitudinal


# noinspection PyPep8Naming
class Wake(object):
    # noinspection PyUnresolvedReferences
    """Class to generate a wake object

    The wake object is defined by its srange, specified
    at initialization, and one or several components corresponding
    to transverse dipoles or quadrupoles, or longitudinal wakes.

    The ``srange`` is common to all components and cannot be changed
    once initialized, all added component are resampled to the
    ``srange``.

    Parameters:
        srange:         vector of s position where to sample the wake

    Example:
        >>> # Build the range
        >>> srange = Wake.build_srange(-0.36, 0.36, 1.0e-5, 1.0e-2,
        ... ring.circumference, ring.circumference)
        >>> # Create the Wake object
        >>> wake = Wake(srange)
        >>> # Build a resonator and add it to the Wake
        >>> freq = 10e9
        >>> Q = 1
        >>> Rs = 1e4
        >>> wake.add(WakeType.RESONATOR, WakeComponent.Z, freq, Q, Rs,
        ... ring.beta)

        See :py:meth:`Wake.long_resonator` for a simple way of creatong the
        same :py:class:`Wake`

    Components may be retrieved using :code:`wake.DX` for example
    """
    def __init__(self, srange):
        assert len(srange) == len(numpy.unique(srange)), \
            "srange must contain unique values"            
        self._srange = srange
        self.components = {WakeComponent.DX: None,
                           WakeComponent.DY: None,
                           WakeComponent.QX: None,
                           WakeComponent.QY: None,
                           WakeComponent.Z: None}

    @property
    def srange(self):
        return self._srange

    @property
    def DX(self):
        """Dipole X component"""
        return self.components[WakeComponent.DX]

    @property
    def DY(self):
        """Dipole Y component"""
        return self.components[WakeComponent.DY]

    @property
    def QX(self):
        """Quadrupole X component"""
        return self.components[WakeComponent.QX]

    @property
    def QY(self):
        """Quadrupole Y component"""
        return self.components[WakeComponent.QY]

    @property
    def Z(self):
        """Longitudinal component"""
        return self.components[WakeComponent.Z]

    def add(self, wtype: WakeType, wcomp: WakeComponent, *args, **kwargs):
        """Add a component to a :py:class:`.Wake`

        Parameters:
            wtype:      Wake type
            wcomp:      Wake component
            *args:      parameters for the selected component
        """
        if wtype is WakeType.FILE:
            w = self._readwakefile(*args, **kwargs)
        elif wtype is WakeType.TABLE:
            w = self._resample(*args)
        elif wtype is WakeType.RESONATOR:
            w = self._resonator(wcomp, *args, **kwargs)
        elif wtype is WakeType.RESWALL:
            w = self._reswall(wcomp, *args, **kwargs)

        else:
            raise AtError('Invalid WakeType: {}'.format(wtype))
        if self.components[wcomp] is None:
            self.components[wcomp] = w
        else:
            self.components[wcomp] += w

    def _resample(self, s, w):
        if self._srange[0] < s[0] or self._srange[-1] < s[-1]:
            warnings.warn(AtWarning('Input wake is smaller '
                                    'than desired Wake() range. '
                                    'Filling with zeros.\n'))
        fint = interp1d(s, w, bounds_error=False, fill_value=0)
        wint = fint(self._srange)
        return wint

    def _readwakefile(self, filename, scol=0, wcol=1, sfact=1, wfact=1,
                      delimiter=None, skiprows=0):
        s, w = numpy.loadtxt(filename, delimiter=delimiter, unpack=True,
                             usecols=(scol, wcol), skiprows=skiprows)
        s *= sfact
        w *= wfact
        return self._resample(s, w)

    def _resonator(self, wcomp, frequency, qfactor, rshunt, beta,
                   yokoya_factor=1):
        
        if numpy.amin(self._srange) < 0:
            raise ValueError("""
                            Provided srange has negative values.
                            This is not allowed for the resonator
                            wake function. Please correct.
                            """)

        # It is needed to have a point at 0 and 1e-24 to sample properly
        # the discontinuity in the longitudinal plane

        self._srange = numpy.unique(numpy.concatenate(([0.0, 1e-24],
                                                        self._srange)))

        if wcomp is WakeComponent.Z:
            return long_resonator_wf(self._srange, frequency,
                                     qfactor, rshunt, beta)
        elif isinstance(wcomp, WakeComponent):
            return transverse_resonator_wf(self._srange, frequency,
                                           qfactor, rshunt, yokoya_factor,
                                           beta)
        else:
            raise AtError('Invalid WakeComponent: {}'.format(wcomp))

    def _reswall(self, wcomp, length, rvac, conduct, beta, yokoya_factor=1):
    
        if numpy.amin(self._srange) < 0:
            raise ValueError("""
                            Provided srange has negative values.
                            This is not allowed for the resistive wall
                            wake function. Please correct.
                            """)

        # It is needed to have a point at 0 and 1e-24 to sample properly
        # the discontinuity in the longitudinal plane (if combining with
        # longres)
        self._srange = numpy.unique(numpy.concatenate(([0.0, 1e-24],
                                                       self._srange)))

        if wcomp is WakeComponent.Z:
            raise AtError('Resitive wall not available '
                          'for WakeComponent: {}'.format(wcomp))
        elif isinstance(wcomp, WakeComponent):
            return transverse_reswall_wf(self._srange, yokoya_factor,
                                         length, rvac, conduct, beta)
        else:
            raise AtError('Invalid WakeComponent: {}'.format(wcomp))

    @staticmethod
    def resonator(srange, wakecomp,
                  frequency, qfactor, rshunt,
                  beta: float, yokoya_factor=1, nelems: int = 1):
        r"""Factory method to build a resonator wake object

        Parameters:
            srange:         vector of s position where to sample the wake
            wakecomp:       Wake component
            frequency:      Resonator frequency [Hz]
            qfactor:        Q factor
            rshunt:         Shunt impedance, [:math:`\Omega`] for longitudinal,
              [:math:`\Omega/m`] for transverse
            beta:           Relativistic :math:`\beta`
            yokoya_factor:  Yokoya factor
            nelems:         Number of resonators

        Several resonators can be built in one step by setting ``nelems``> 1
        and giving array parameters
        """
        wake = Wake(srange)
        try:
            wakecomp = numpy.broadcast_to(wakecomp, (nelems, ))
            frequency = numpy.broadcast_to(frequency, (nelems, ))
            qfactor = numpy.broadcast_to(qfactor, (nelems, ))
            rshunt = numpy.broadcast_to(rshunt, (nelems, ))
            yokoya_factor = numpy.broadcast_to(yokoya_factor, (nelems, ))
        except ValueError:
            raise AtError('Wake object inputs should be either scalars '
                          'or with shape (len(wakecomp), )')
        for wc, fr, qf, rs, yk in zip(wakecomp, frequency, qfactor, rshunt,
                                      yokoya_factor):
            wake.add(WakeType.RESONATOR, wc, fr, qf, rs, beta,
                     yokoya_factor=yk)
        return wake

    @staticmethod
    def long_resonator(srange, frequency, qfactor, rshunt, beta, nelems=1):
        # noinspection PyUnresolvedReferences
        r"""Factory method to build a longitudinal resonator wake object

        Parameters:
            srange:         vector of s position where to sample the wake
            frequency:      Resonator frequency [Hz]
            qfactor:        Q factor
            rshunt:         Shunt impedance, [:math:`\Omega`] for longitudinal,
              [:math:`\Omega/m`] for transverse
            beta:           Relativistic :math:`\beta`
            nelems:         Number of resonators

        Several resonators can be built in one step by setting ``nelems``> 1
        and giving array parameters

        Example:
            >>> # Build the range
            >>> srange = Wake.build_srange(-0.36, 0.36, 1.0e-5, 1.0e-2,
            ... ring.circumference, ring.circumference)
            >>> # Create the resonator
            >>> freq = 10e9
            >>> Q = 1
            >>> Rs = 1e4
            >>> wake = Wake.long_resonator(srange, freq, Q, Rs, ring.beta)
        """ 
        return Wake.resonator(srange, WakeComponent.Z, frequency, qfactor,
                              rshunt, beta, nelems=nelems)

    @staticmethod
    def resistive_wall(srange, wakecomp: WakeComponent,
                       length, rvac, conduct, beta: float,
                       yokoya_factor=1, nelems: int = 1):
        r"""Factory method to build a resistive wall wake object

        Parameters:
            srange:         vector of s position where to sample the wake
            wakecomp:       Wake component
            length:
            rvac:
            conduct:
            beta:           Relativistic :math:`\beta`
            yokoya_factor:  Yokoya factor
            nelems:
        """
        wake = Wake(srange)
        try:
            wakecomp = numpy.broadcast_to(wakecomp, (nelems, ))
            length = numpy.broadcast_to(length, (nelems, ))
            rvac = numpy.broadcast_to(rvac, (nelems, ))
            conduct = numpy.broadcast_to(conduct, (nelems, ))
            yokoya_factor = numpy.broadcast_to(yokoya_factor, (nelems, ))
        except ValueError:
            raise AtError('Wake object inputs should be either scalars '
                          'or with shape (len(wakecomp), )')
        for wc, le, rv, co, yk in zip(wakecomp, length, rvac,
                                      conduct, yokoya_factor):
            wake.add(WakeType.RESWALL, wc, le, rv, co, beta, yokoya_factor=yk)
        return wake

    @staticmethod
    def build_srange(start: float, bunch_ext: float, short_step: float,
                     long_step: float,
                     bunch_interval: float, totallength: float):
        """Function to build the wake table s column.
        This is not the slicing but the look-up table,
        however it generates data where bunches are located
        to avoid using too much memory to store the table.

        Parameters:
            start:          starting s-coordinate of the table
                            (can be negative for wake potential)
            bunch_ext:      maximum bunch extension, function
                            generates data at +/- bunch_ext
                            around the bucket center
            short_step:     step size for the short range wake table
            long_step:      step size for the long range wake table
            bunch_interval: minimum bunch interval. Data will be generated
                            for each ``bunch_interval`` step
            totallength:    total length of the wake table, has to contain
                            the full bunch extension

        Returns:
            srange:         vector of s position where to sample the wake
        """
        srange = numpy.arange(start, bunch_ext, short_step)
        rangel = numpy.arange(-bunch_ext, bunch_ext, long_step)
        nbunch = int((totallength-bunch_ext)/bunch_interval)
        for i in range(nbunch):
            srange = numpy.concatenate((srange, rangel+bunch_interval*(i+1)))
        return numpy.unique(srange)



from enum import Enum
from warnings import warn
from math import pi
from typing import Optional, Tuple
import numpy
from scipy.optimize import least_squares
from at.lattice import Lattice, Dipole, Wiggler, RFCavity, Refpts, EnergyLoss
from at.lattice import check_radiation, AtError, AtWarning
from at.lattice import QuantumDiffusion, Collective, SimpleQuantDiff
from at.lattice import get_bool_index, set_value_refpts
from at.constants import clight, Cgamma
from at.tracking import internal_lpass

__all__ = ['get_energy_loss', 'set_cavity_phase', 'ELossMethod',
           'get_timelag_fromU0']


class ELossMethod(Enum):
    """methods for the computation of energy losses"""
    #: The losses are obtained from
    #: :math:`E_{loss}=C_\gamma/2\pi . E^4 . I_2`.
    #: Takes into account bending magnets and wigglers.
    INTEGRAL = 1
    #: The losses are obtained by tracking without cavities.
    #: Needs radiation ON, takes into account all radiating elements
    TRACKING = 2


def get_energy_loss(ring: Lattice,
                    method: Optional[ELossMethod] = ELossMethod.INTEGRAL
                    ) -> float:
    """Computes the energy loss per turn

    Parameters:
        ring:           Lattice description
        method:         Method for energy loss computation.
          See :py:class:`ELossMethod`.

    Returns:
        eloss (float):  Energy loss per turn [eV]
    """

    # noinspection PyShadowingNames
    def integral(ring):
        """Losses = Cgamma / 2pi * EGeV^4 * i2
        """

        def wiggler_i2(wiggler: Wiggler):
            rhoinv = wiggler.Bmax / ring.BRho
            coefh = wiggler.By[1, :]
            coefv = wiggler.Bx[1, :]
            return wiggler.Length * (numpy.sum(coefh * coefh) + numpy.sum(
                coefv*coefv)) * rhoinv ** 2 / 2

        def dipole_i2(dipole: Dipole):
            return dipole.BendingAngle ** 2 / dipole.Length

        def eloss_i2(eloss: EnergyLoss):
            return eloss.EnergyLoss / coef

        i2 = 0.0
        coef = Cgamma / 2.0 / pi * ring.energy ** 4
        for el in ring:
            if isinstance(el, Dipole):
                i2 += dipole_i2(el)
            elif isinstance(el, Wiggler) and el.PassMethod != 'DriftPass':
                i2 += wiggler_i2(el)
            elif isinstance(el, EnergyLoss) and el.PassMethod != 'IdentityPass':
                i2 += eloss_i2(el)
        e_loss = coef * i2
        return e_loss

    # noinspection PyShadowingNames
    @check_radiation(True)
    def tracking(ring):
        """Losses from tracking
        """
        ringtmp = ring.disable_6d(RFCavity, QuantumDiffusion, Collective,
                                  SimpleQuantDiff, copy=True)
        o6 = numpy.squeeze(internal_lpass(ringtmp, numpy.zeros(6),
                                          refpts=len(ringtmp)))
        if numpy.isnan(o6[0]):
            dp = 0
            for e in ringtmp:
                ot = numpy.squeeze(internal_lpass([e], numpy.zeros(6)))
                dp += -ot[4] * ring.energy
            return dp
        else:
            return -o6[4] * ring.energy

    if isinstance(method, str):
        method = ELossMethod[method.upper()]
        warn(FutureWarning('You should use {0!s}'.format(method)))
    if method is ELossMethod.INTEGRAL:
        return ring.periodicity * integral(ring)
    elif method == ELossMethod.TRACKING:
        return ring.periodicity * tracking(ring)
    else:
        raise AtError('Invalid method: {}'.format(method))


# noinspection PyPep8Naming
def get_timelag_fromU0(ring: Lattice,
                       method: Optional[ELossMethod] = ELossMethod.TRACKING,
                       cavpts: Optional[Refpts] = None,
                       divider: Optional[int] = 4,
                       ts_tol: Optional[float] = 1.0e-9) -> Tuple[float, float]:
    """
    Get the TimeLag attribute of RF cavities based on frequency,
    voltage and energy loss per turn, so that the synchronous phase is zero.
    An error occurs if all cavities do not have the same frequency.
    Used in set_cavity_phase()


    Parameters:
        ring:               Lattice description
        method:             Method for energy loss computation.
          See :py:class:`ELossMethod`.
        cavpts:             Cavity location. If None, use all cavities.
          This allows to ignore harmonic cavities.
        divider: number of segments to search for ts
        phis_tol: relative tolerance for ts calculation
    Returns:
        timelag (float):    Timelag
        ts (float):         Time difference with the present value
    """
    def singlev(values):
        vals = numpy.unique(values)
        if len(vals) > 1:
            raise AtError('values not equal for all cavities')
        return vals[0]

    def eq(x, freq, rfv, tl0, u0):
        omf = 2*numpy.pi*freq/clight
        if u0 > 0.0:
            eq1 = (numpy.sum(-rfv*numpy.sin(omf*(x-tl0)))-u0)/u0
        else:
            eq1 = numpy.sum(-rfv * numpy.sin(omf * (x - tl0)))
        eq2 = numpy.sum(-omf*rfv*numpy.cos(omf*(x-tl0)))
        if eq2 > 0:
            return numpy.sqrt(eq1**2+eq2**2)
        else:
            return abs(eq1)

    if cavpts is None:
        cavpts = get_bool_index(ring, RFCavity)
    u0 = get_energy_loss(ring, method=method) / ring.periodicity
    freq = numpy.array([cav.Frequency for cav in ring.select(cavpts)])
    rfv = numpy.array([cav.Voltage for cav in ring.select(cavpts)])
    tl0 = numpy.array([cav.TimeLag for cav in ring.select(cavpts)])
    try:
        frf = singlev(freq)
        tml = singlev(tl0)
    except AtError:
        ctmax = clight/numpy.amin(freq)/2
        tt0 = tl0[numpy.argmin(freq)]
        bounds = (-ctmax, ctmax)
        args = (freq, rfv, tl0, u0)
        r = []
        for i in range(divider):
            fact = (i+1)/divider
            r.append(least_squares(eq, bounds[0]*fact+tt0,
                                   args=args, bounds=bounds+tt0))
            r.append(least_squares(eq, bounds[1]*fact+tt0,
                                   args=args, bounds=bounds+tt0))
        res = numpy.array([ri.fun[0] for ri in r])
        ok = res < ts_tol
        vals = numpy.array([abs(ri.x[0]).round(decimals=6) for ri in r])
        if not numpy.any(ok):
            raise AtError('No solution found for Phis, please check '
                          'RF settings')
        if len(numpy.unique(vals[ok])) > 1:
            warn(AtWarning('More than one solution found for Phis: use '
                           'best fit, please check RF settings'))
        ts = -r[numpy.argmin(res)].x[0]
        timelag = ts+tl0
    else:
        if u0 > numpy.sum(rfv):
            raise AtError('Not enough RF voltage: unstable ring')
        vrf = numpy.sum(rfv)
        timelag = clight/(2*numpy.pi*frf)*numpy.arcsin(u0/vrf)
        ts = timelag - tml
        timelag *= numpy.ones(ring.refcount(cavpts))
    return timelag, ts


def set_cavity_phase(ring: Lattice,
                     method: ELossMethod = ELossMethod.TRACKING,
                     refpts: Optional[Refpts] = None,
                     cavpts: Optional[Refpts] = None,
                     copy: bool = False) -> None:
    """
    Adjust the TimeLag attribute of RF cavities based on frequency,
    voltage and energy loss per turn, so that the synchronous phase is zero.
    An error occurs if all cavities do not have the same frequency.

    .. Warning::

       This function changes the time reference, this should be avoided

    Parameters:
        ring:       Lattice description
        method:     Method for energy loss computation.
          See :py:class:`ELossMethod`.
        cavpts:     Cavity location. If None, use all cavities.
          This allows to ignore harmonic cavities.
        copy:       If True, returns a shallow copy of ring with new
          cavity elements. Otherwise, modify ring in-place.
    """
    # refpts is kept for backward compatibility
    if cavpts is None and refpts is not None:
        warn(FutureWarning('You should use "cavpts" instead of "refpts"'))
        cavpts = refpts
    elif cavpts is None:
        cavpts = get_bool_index(ring, RFCavity)
    timelag, _ = get_timelag_fromU0(ring, method=method, cavpts=cavpts)
    set_value_refpts(ring, cavpts, 'TimeLag', timelag, copy=copy)


Lattice.get_energy_loss = get_energy_loss
Lattice.energy_loss = property(get_energy_loss)
Lattice.set_cavity_phase = set_cavity_phase
